#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Source classes."""

from enum import Enum
from types import BuiltinFunctionType, FunctionType, ModuleType
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Type, Union
from uuid import UUID

from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    PlainSerializer,
    SerializeAsAny,
    field_validator,
)
from typing_extensions import Annotated

from zenml.logger import get_logger

if TYPE_CHECKING:
    AnyClassMethod = classmethod[Any]  # type: ignore[type-arg]

logger = get_logger(__name__)


class SourceType(Enum):
    """Enum representing different types of sources."""

    USER = "user"
    BUILTIN = "builtin"
    INTERNAL = "internal"
    DISTRIBUTION_PACKAGE = "distribution_package"
    CODE_REPOSITORY = "code_repository"
    NOTEBOOK = "notebook"
    UNKNOWN = "unknown"


class Source(BaseModel):
    """Source specification.

    A source specifies a module name as well as an optional attribute of that
    module. These values can be used to import the module and get the value
    of the attribute inside the module.

    Example:
        The source `Source(module="zenml.config.source", attribute="Source")`
        references the class that this docstring is describing. This class is
        defined in the `zenml.config.source` module and the name of the
        attribute is the class name `Source`.

    Attributes:
        module: The module name.
        attribute: Optional name of the attribute inside the module.
        type: The type of the source.
    """

    module: str
    attribute: Optional[str] = None
    type: SourceType

    @classmethod
    def from_import_path(
        cls, import_path: str, is_module_path: bool = False
    ) -> "Source":
        """Creates a source from an import path.

        Args:
            import_path: The import path.
            is_module_path: If the import path points to a module or not.

        Raises:
            ValueError: If the import path is empty.

        Returns:
            The source.
        """
        if not import_path:
            raise ValueError(
                "Invalid empty import path. The import path needs to refer "
                "to a Python module and an optional attribute of that module."
            )

        # Remove internal version pins for backwards compatibility
        if "@" in import_path:
            import_path = import_path.split("@", 1)[0]

        if is_module_path or "." not in import_path:
            module = import_path
            attribute = None
        else:
            module, attribute = import_path.rsplit(".", maxsplit=1)

        return Source(
            module=module, attribute=attribute, type=SourceType.UNKNOWN
        )

    @property
    def import_path(self) -> str:
        """The import path of the source.

        Returns:
            The import path of the source.
        """
        if self.attribute:
            return f"{self.module}.{self.attribute}"
        else:
            return self.module

    @property
    def is_internal(self) -> bool:
        """If the source is internal (=from the zenml package).

        Returns:
            True if the source is internal, False otherwise
        """
        if self.type not in {SourceType.UNKNOWN, SourceType.INTERNAL}:
            return False

        return self.module.split(".", maxsplit=1)[0] == "zenml"

    @property
    def is_module_source(self) -> bool:
        """If the source is a module source.

        Returns:
            If the source is a module source.
        """
        return self.attribute is None

    model_config = ConfigDict(extra="allow")

    def model_dump(self, **kwargs: Any) -> Dict[str, Any]:
        """Dump the source as a dictionary.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            The source as a dictionary.
        """
        return super().model_dump(serialize_as_any=True, **kwargs)

    def model_dump_json(self, **kwargs: Any) -> str:
        """Dump the source as a JSON string.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            The source as a JSON string.
        """
        return super().model_dump_json(serialize_as_any=True, **kwargs)

    @classmethod
    def convert_source(cls, source: Any) -> Any:
        """Converts an old source string to a source object.

        Args:
            source: Source string or object.

        Returns:
            The converted source.
        """
        if isinstance(source, str):
            source = cls.from_import_path(source)

        return source


ObjectType = Union[
    Type[Any],
    Callable[..., Any],
    ModuleType,
    FunctionType,
    BuiltinFunctionType,
]


class SourceOrObject(Source):
    """Hybrid type that can hold either a Source path or a loaded object (type, function, variable, etc.).

    This enables:
    - Internal use: Pass actual objects directly
    - External use: Pass source strings from config
    - Lazy serialization: Converts objects to source path strings only at
    serialization time
    - Lazy loading: Only loads sources when explicitly requested

    Examples:
    * the source `SourceOrObject(module="zenml.config.source", attribute="SourceOrObject")`
    references the class that this docstring is describing.
    * the object `SourceOrObject.from_object(object=my_object)` creates a source
    or object from the object `my_object`.

    Attributes:
        _object: Loaded object object (if initialized from object or loaded).
        _is_loaded: Whether the callable has been loaded.
    """

    _object: Optional[ObjectType] = None
    _is_loaded: bool = False

    @classmethod
    def from_source(cls, source: Source) -> "SourceOrObject":
        """Creates a source or object from a source instance.

        Args:
            source: The source instance.

        Returns:
            The source or object instance.
        """
        return cls(**source.model_dump())

    @classmethod
    def from_object(cls, object: ObjectType) -> "SourceOrObject":
        """Creates a source or object from an object instance.

        Args:
            object: The object instance.

        Returns:
            The source or object instance.
        """
        # We don't resolve the object right away, we only do it at serialization
        # time. This allows us to store temporary objects in this class too.
        result = cls(
            module="",
            type=SourceType.UNKNOWN,
        )
        result._object = object
        result._is_loaded = True
        return result

    def load(self) -> ObjectType:
        """Load and return the object.

        Returns:
            The loaded object.

        Raises:
            RuntimeError: If loading fails.
        """
        from zenml.utils import source_utils

        if not self._is_loaded:
            try:
                self._object = source_utils.load(self)
                self._is_loaded = True
            except Exception as e:
                raise RuntimeError(
                    f"Failed to load object from {self.import_path}: {e}"
                ) from e

        assert self._object is not None
        return self._object

    def resolve(self) -> None:
        """Resolve the stored object to a source."""
        from zenml.utils import source_utils

        if not self._is_loaded:
            return

        if self.module:
            # Already resolved
            return

        source = source_utils.resolve(self._object)
        for k, v in source.model_dump().items():
            setattr(self, k, v)

    @property
    def is_loaded(self) -> bool:
        """Whether the object has been loaded.

        Returns:
            True if the object has been loaded, False otherwise.
        """
        return self._is_loaded

    @classmethod
    def convert_source_or_object(
        cls,
        source: Union[
            str, "SourceOrObject", Source, Dict[str, Any], ObjectType
        ],
    ) -> Union["SourceOrObject", Dict[str, Any]]:
        """Converts a source string or object to a SourceOrObject object.

        Args:
            source: Source string or object.

        Returns:
            The converted source or object.
        """
        if isinstance(source, str):
            source = cls.from_import_path(source)

        if isinstance(source, cls):
            return source

        if isinstance(source, Source):
            return cls.from_source(source)

        if isinstance(source, dict):
            return source

        return cls.from_object(source)

    def model_dump(self, **kwargs: Any) -> Dict[str, Any]:
        """Dump the source as a dictionary.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            The source as a dictionary.
        """
        self.resolve()
        return super().model_dump(**kwargs)

    def model_dump_json(self, **kwargs: Any) -> str:
        """Dump the source as a JSON string.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            The source as a JSON string.
        """
        self.resolve()
        return super().model_dump_json(**kwargs)

    @classmethod
    def serialize_source_or_object(
        cls, value: "SourceOrObject"
    ) -> Dict[str, Any]:
        """Serialize the source or object as a dictionary.

        Args:
            value: The source or object to serialize.

        Returns:
            The source or object as a string.
        """
        return value.model_dump()


class DistributionPackageSource(Source):
    """Source representing an object from a distribution package.

    Attributes:
        package_name: Name of the package.
        version: The package version.
    """

    package_name: str
    version: Optional[str] = None
    type: SourceType = SourceType.DISTRIBUTION_PACKAGE

    @field_validator("type")
    @classmethod
    def _validate_type(cls, value: SourceType) -> SourceType:
        """Validate the source type.

        Args:
            value: The source type.

        Raises:
            ValueError: If the source type is not `DISTRIBUTION_PACKAGE`.

        Returns:
            The source type.
        """
        if value != SourceType.DISTRIBUTION_PACKAGE:
            raise ValueError("Invalid source type.")

        return value


class CodeRepositorySource(Source):
    """Source representing an object from a code repository.

    Attributes:
        repository_id: The code repository ID.
        commit: The commit.
        subdirectory: The subdirectory of the source root inside the code
            repository.
    """

    repository_id: UUID
    commit: str
    subdirectory: str
    type: SourceType = SourceType.CODE_REPOSITORY

    @field_validator("type")
    @classmethod
    def _validate_type(cls, value: SourceType) -> SourceType:
        """Validate the source type.

        Args:
            value: The source type.

        Raises:
            ValueError: If the source type is not `CODE_REPOSITORY`.

        Returns:
            The source type.
        """
        if value != SourceType.CODE_REPOSITORY:
            raise ValueError("Invalid source type.")

        return value


class NotebookSource(Source):
    """Source representing an object defined in a notebook.

    Attributes:
        replacement_module: Name of the module from which this source should
            be loaded in case the code is not running in a notebook.
        artifact_store_id: ID of the artifact store in which the replacement
            module code is stored.
    """

    replacement_module: Optional[str] = None
    artifact_store_id: Optional[UUID] = None
    type: SourceType = SourceType.NOTEBOOK

    @field_validator("type")
    @classmethod
    def _validate_type(cls, value: SourceType) -> SourceType:
        """Validate the source type.

        Args:
            value: The source type.

        Raises:
            ValueError: If the source type is not `NOTEBOOK`.

        Returns:
            The source type.
        """
        if value != SourceType.NOTEBOOK:
            raise ValueError("Invalid source type.")

        return value

    @field_validator("module")
    @classmethod
    def _validate_module(cls, value: str) -> str:
        """Validate the module.

        Args:
            value: The module.

        Raises:
            ValueError: If the module is not `__main__`.

        Returns:
            The module.
        """
        if value != "__main__":
            raise ValueError("Invalid module for notebook source.")

        return value


SourceWithValidator = Annotated[
    SerializeAsAny[Source],
    BeforeValidator(Source.convert_source),
]

if TYPE_CHECKING:
    SourceOrObjectField = Union[ObjectType, SourceOrObject, Source, str]
else:
    SourceOrObjectField = Annotated[
        SourceOrObject,
        BeforeValidator(SourceOrObject.convert_source_or_object),
        PlainSerializer(
            SourceOrObject.serialize_source_or_object, return_type=dict
        ),
    ]



"""Release [0.56.2].

Revision ID: 0.56.2
Revises: 0701da9951a0
Create Date: 2024-03-25 14:49:49.021147

"""

# revision identifiers, used by Alembic.
revision = "0.56.2"
down_revision = "0701da9951a0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Framework adapter interfaces."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, cast

from asgiref.typing import (
    ASGIApplication,
    ASGIReceiveCallable,
    ASGISendCallable,
    Scope,
)

from zenml.config.source import SourceOrObject

if TYPE_CHECKING:
    from zenml.config import (
        EndpointSpec,
        MiddlewareSpec,
    )
    from zenml.deployers.server.app import BaseDeploymentAppRunner


class EndpointAdapter(ABC):
    """Converts framework-agnostic endpoint specs to framework endpoints."""

    def resolve_endpoint_handler(
        self,
        app_runner: "BaseDeploymentAppRunner",
        endpoint_spec: "EndpointSpec",
    ) -> Any:
        """Resolve an endpoint handler from its specification.

        This method handles three types of handlers as defined in EndpointSpec:
        1. Direct endpoint function - returned as-is
        2. Endpoint builder class - instantiated with app_runner, app, and
        init_kwargs
        3. Endpoint builder function - called with app_runner, app, and
        init_kwargs to obtain the actual endpoint

        Args:
            app_runner: Deployment app runner instance.
            endpoint_spec: The endpoint specification to resolve the handler
                from.

        Returns:
            The actual endpoint callable ready to be registered.

        Raises:
            ValueError: If handler is not callable or builder returns
                non-callable.
            RuntimeError: If handler resolution fails.
        """
        import inspect

        assert isinstance(endpoint_spec.handler, SourceOrObject)
        handler = endpoint_spec.handler.load()

        if endpoint_spec.native:
            return handler

        # Type 2: Endpoint builder class
        if isinstance(handler, type):
            if not hasattr(handler, "__call__"):
                raise ValueError(
                    f"Handler class {handler.__name__} must implement "
                    "__call__ method"
                )
            try:
                inner_handler = handler(
                    app_runner=app_runner,
                    **endpoint_spec.init_kwargs,
                )
            except TypeError as e:
                raise RuntimeError(
                    f"Failed to instantiate handler class "
                    f"{handler.__name__}: {e}"
                ) from e

            if not callable(inner_handler):
                raise ValueError(
                    f"The __call__ method of the handler class "
                    f"{handler.__name__} must return a callable"
                )

            return inner_handler

        if not callable(handler):
            raise ValueError(f"Handler {handler} is not callable")

        # Determine if it's Type 3 (builder function) or Type 1 (direct)
        try:
            sig = inspect.signature(handler)
            params = set(sig.parameters.keys())

            # Type 3: Builder function (has app_runner parameter)
            if "app_runner" in params:
                try:
                    inner_handler = handler(
                        app_runner=app_runner,
                        **endpoint_spec.init_kwargs,
                    )
                    if not callable(inner_handler):
                        raise ValueError(
                            f"Builder function {handler.__name__} must "
                            f"return a callable, got {type(inner_handler)}"
                        )
                    return inner_handler
                except TypeError as e:
                    raise RuntimeError(
                        f"Failed to call builder function "
                        f"{handler.__name__}: {e}"
                    ) from e

            # Type 1: Direct endpoint function
            return handler

        except ValueError:
            # inspect.signature failed, assume it's a direct endpoint
            return handler

    @abstractmethod
    def register_endpoint(
        self,
        app_runner: "BaseDeploymentAppRunner",
        spec: "EndpointSpec",
    ) -> None:
        """Register an endpoint on the app.

        Args:
            app_runner: Deployment app runner instance.
            spec: Framework-agnostic endpoint specification.

        Raises:
            RuntimeError: If endpoint registration fails.
        """


class MiddlewareAdapter(ABC):
    """Converts framework-agnostic middleware specs to framework middleware."""

    def resolve_middleware_handler(
        self,
        app_runner: "BaseDeploymentAppRunner",
        middleware_spec: "MiddlewareSpec",
    ) -> Any:
        """Resolve a middleware handler from its specification.

        This method handles three types of middleware as defined in MiddlewareSpec:
        1. Middleware callable class
        2. Middleware callable function
        3. Native middleware object

        Args:
            app_runner: Deployment app runner instance.
            middleware_spec: The middleware specification to resolve the handler
                from.

        Returns:
            The actual middleware callable ready to be registered.

        Raises:
            ValueError: If middleware is not callable or builder returns
                non-callable.
        """
        import inspect

        assert isinstance(middleware_spec.middleware, SourceOrObject)
        middleware = middleware_spec.middleware.load()

        if middleware_spec.native:
            return middleware

        # Type 1: Middleware class
        if isinstance(middleware, type):
            return middleware

        if not callable(middleware):
            raise ValueError(f"Middleware {middleware} is not callable")

        # Wrap the middleware function in a middleware class
        class _MiddlewareAdapter:
            def __init__(self, app: ASGIApplication, **kwargs: Any) -> None:
                self.app = app
                self.kwargs = kwargs

            async def __call__(
                self,
                scope: Scope,
                receive: ASGIReceiveCallable,
                send: ASGISendCallable,
            ) -> None:
                callable_middleware = cast(Callable[..., Any], middleware)
                if inspect.iscoroutinefunction(callable_middleware):
                    await callable_middleware(
                        app=self.app,
                        scope=scope,
                        receive=receive,
                        send=send,
                        **self.kwargs,
                    )
                else:
                    callable_middleware(
                        app=self.app,
                        scope=scope,
                        receive=receive,
                        send=send,
                        **self.kwargs,
                    )

        return _MiddlewareAdapter

    @abstractmethod
    def register_middleware(
        self,
        app_runner: "BaseDeploymentAppRunner",
        spec: "MiddlewareSpec",
    ) -> None:
        """Register middleware on the app.

        Args:
            app_runner: Deployment app runner instance.
            spec: Framework-agnostic middleware specification.

        Raises:
            ValueError: If middleware scope requires missing parameters.
            RuntimeError: If middleware registration fails.
        """



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline run DAG models."""

from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import BaseModel

from zenml.enums import ExecutionStatus


class PipelineRunDAG(BaseModel):
    """Pipeline run DAG."""

    id: UUID
    status: ExecutionStatus
    nodes: List["Node"]
    edges: List["Edge"]

    class Node(BaseModel):
        """Node in the pipeline run DAG."""

        node_id: str
        type: str
        id: Optional[UUID] = None
        name: str
        metadata: Dict[str, Any] = {}

    class Edge(BaseModel):
        """Edge in the pipeline run DAG."""

        source: str
        target: str
        metadata: Dict[str, Any] = {}



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Databricks orchestrator base config and settings."""

from typing import TYPE_CHECKING, Dict, Optional, Tuple, Type

from pydantic import Field

from zenml.config.base_settings import BaseSettings
from zenml.integrations.databricks import DATABRICKS_ORCHESTRATOR_FLAVOR
from zenml.logger import get_logger
from zenml.orchestrators import BaseOrchestratorConfig
from zenml.orchestrators.base_orchestrator import BaseOrchestratorFlavor
from zenml.utils.enum_utils import StrEnum
from zenml.utils.secret_utils import SecretField

if TYPE_CHECKING:
    from zenml.integrations.databricks.orchestrators import (
        DatabricksOrchestrator,
    )

logger = get_logger(__name__)


class DatabricksAvailabilityType(StrEnum):
    """Databricks availability type."""

    ON_DEMAND = "ON_DEMAND"
    SPOT = "SPOT"
    SPOT_WITH_FALLBACK = "SPOT_WITH_FALLBACK"


class DatabricksOrchestratorSettings(BaseSettings):
    """Databricks orchestrator base settings.

    Configuration for Databricks cluster and Spark execution settings.
    Field descriptions are defined inline using Field() descriptors.
    """

    # Cluster Configuration
    spark_version: Optional[str] = Field(
        default=None,
        description="Apache Spark version for the Databricks cluster. "
        "Uses workspace default if not specified. Example: '16.4.x-scala2.12'",
    )
    num_workers: Optional[int] = Field(
        default=None,
        description="Fixed number of worker nodes. Cannot be used with autoscaling.",
    )
    node_type_id: Optional[str] = Field(
        default=None,
        description="Databricks node type identifier. "
        "Refer to Databricks documentation for available instance types. "
        "Example: 'i3.xlarge'",
    )
    policy_id: Optional[str] = Field(
        default=None,
        description="Databricks cluster policy ID for governance and cost control.",
    )
    autotermination_minutes: Optional[int] = Field(
        default=None,
        description="Minutes of inactivity before automatic cluster termination. "
        "Helps control costs by shutting down idle clusters.",
    )
    autoscale: Tuple[int, int] = Field(
        default=(0, 1),
        description="Cluster autoscaling bounds as (min_workers, max_workers). "
        "Automatically adjusts cluster size based on workload.",
    )
    single_user_name: Optional[str] = Field(
        default=None,
        description="Databricks username for single-user cluster access mode.",
    )
    spark_conf: Optional[Dict[str, str]] = Field(
        default=None,
        description="Custom Spark configuration properties as key-value pairs. "
        "Example: {'spark.sql.adaptive.enabled': 'true', 'spark.sql.adaptive.coalescePartitions.enabled': 'true'}",
    )
    spark_env_vars: Optional[Dict[str, str]] = Field(
        default=None,
        description="Environment variables for the Spark driver and executors. "
        "Example: {'SPARK_WORKER_MEMORY': '4g', 'SPARK_DRIVER_MEMORY': '2g'}",
    )
    schedule_timezone: Optional[str] = Field(
        default=None,
        description="Timezone for scheduled pipeline execution. "
        "Uses IANA timezone format (e.g., 'America/New_York').",
    )
    availability_type: Optional[DatabricksAvailabilityType] = Field(
        default=None,
        description="Instance availability type: ON_DEMAND (guaranteed), SPOT (cost-optimized), "
        "or SPOT_WITH_FALLBACK (spot with on-demand backup).",
    )


class DatabricksOrchestratorConfig(
    BaseOrchestratorConfig, DatabricksOrchestratorSettings
):
    """Databricks orchestrator base config.

    Attributes:
        host: Databricks host.
        client_id: Databricks client id.
        client_secret: Databricks client secret.
    """

    host: str
    client_id: str = SecretField(default=None)
    client_secret: str = SecretField(default=None)

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return False

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        Returns:
            True if this config is for a remote component, False otherwise.
        """
        return True

    @property
    def is_schedulable(self) -> bool:
        """Whether the orchestrator is schedulable or not.

        Returns:
            Whether the orchestrator is schedulable or not.
        """
        return True


class DatabricksOrchestratorFlavor(BaseOrchestratorFlavor):
    """Databricks orchestrator flavor."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return DATABRICKS_ORCHESTRATOR_FLAVOR

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/orchestrator/databricks.png"

    @property
    def config_class(self) -> Type[DatabricksOrchestratorConfig]:
        """Returns `KubeflowOrchestratorConfig` config class.

        Returns:
                The config class.
        """
        return DatabricksOrchestratorConfig

    @property
    def implementation_class(self) -> Type["DatabricksOrchestrator"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.databricks.orchestrators import (
            DatabricksOrchestrator,
        )

        return DatabricksOrchestrator



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Dynamic pipeline runner."""

import copy
import inspect
import itertools
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import nullcontext
from datetime import timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ContextManager,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)
from uuid import uuid4

from pydantic import TypeAdapter

from zenml import ExternalArtifact
from zenml.artifacts.in_memory_cache import InMemoryArtifactCache
from zenml.client import Client
from zenml.config.compiler import Compiler
from zenml.config.step_configurations import (
    GroupInfo,
    Step,
    StepConfiguration,
    StepConfigurationUpdate,
)
from zenml.constants import (
    ENV_ZENML_DYNAMIC_PIPELINE_MONITORING_DELAY,
    ENV_ZENML_DYNAMIC_PIPELINE_MONITORING_INTERVAL,
    ENV_ZENML_DYNAMIC_PIPELINE_WORKER_COUNT,
    handle_float_env_var,
    handle_int_env_var,
)
from zenml.enums import (
    ExecutionMode,
    ExecutionStatus,
    GroupType,
    RunWaitConditionLeaseMode,
    RunWaitConditionResolution,
    RunWaitConditionStatus,
    RunWaitConditionType,
    StepRuntime,
)
from zenml.execution.pipeline.dynamic.interactive_input_utils import (
    maybe_enable_interactive_wait_prompt,
    poll_interactive_wait_condition_input,
)
from zenml.execution.pipeline.dynamic.outputs import (
    AnyStepFuture,
    ArtifactFuture,
    BaseFuture,
    BaseStepFuture,
    MapResultsFuture,
    OutputArtifact,
    StepFuture,
    StepRunOutputs,
    _InlineStepFuture,
    _IsolatedStepFuture,
)
from zenml.execution.pipeline.dynamic.run_context import (
    DynamicPipelineRunContext,
)
from zenml.execution.pipeline.dynamic.utils import (
    _Unmapped,
    load_step_run_outputs,
)
from zenml.execution.pipeline.utils import compute_invocation_id
from zenml.execution.step.utils import launch_step
from zenml.logger import get_logger
from zenml.models import (
    ArtifactVersionResponse,
    PipelineRunResponse,
    PipelineRunUpdate,
    PipelineSnapshotResponse,
    RunWaitConditionLeaseUpdate,
    RunWaitConditionRequest,
    RunWaitConditionResolveRequest,
    RunWaitConditionResponse,
    StepRunResponse,
)
from zenml.orchestrators.publish_utils import (
    publish_failed_pipeline_run,
    publish_failed_step_run,
    publish_stopped_step_run,
    publish_successful_pipeline_run,
)
from zenml.pipelines.dynamic.pipeline_definition import DynamicPipeline
from zenml.pipelines.run_utils import create_placeholder_run
from zenml.stack import Stack
from zenml.steps.entrypoint_function_utils import StepArtifact
from zenml.steps.step_invocation import StepInvocation
from zenml.steps.utils import OutputSignature
from zenml.utils import (
    context_utils,
    env_utils,
    exception_utils,
    pydantic_utils,
    source_utils,
    string_utils,
)
from zenml.utils.logging_utils import (
    is_pipeline_logging_enabled,
    setup_logging_context,
)

if TYPE_CHECKING:
    from zenml.config import DockerSettings
    from zenml.orchestrators import BaseOrchestrator
    from zenml.steps import BaseStep


logger = get_logger(__name__)

T = TypeVar("T")


class _WaitConditionPollTimeout(Exception):
    """Raised when a wait condition polling times out."""


class _WaitConditionAborted(Exception):
    """Raised when a wait condition is aborted."""


class _WaitConditionState(NamedTuple):
    """State tuple for wait condition handling."""

    is_terminal: bool
    value: Any


class DynamicPipelineRunner:
    """Dynamic pipeline runner."""

    def __init__(
        self,
        snapshot: "PipelineSnapshotResponse",
        run: Optional["PipelineRunResponse"],
        orchestrator: Optional["BaseOrchestrator"] = None,
    ) -> None:
        """Initialize the dynamic pipeline runner.

        Args:
            snapshot: The snapshot of the pipeline.
            run: The pipeline run.
            orchestrator: The orchestrator to use. If not provided, the
                orchestrator will be inferred from the snapshot stack.

        Raises:
            RuntimeError: If the snapshot has no associated stack.
        """
        if not snapshot.stack:
            raise RuntimeError("Missing stack for snapshot.")

        if (
            snapshot.pipeline_configuration.execution_mode
            == ExecutionMode.CONTINUE_ON_FAILURE
        ):
            logger.warning(
                "The `%s` execution mode is not supported for "
                "dynamic pipelines right now. "
                "The `%s` execution mode will be used instead.",
                snapshot.pipeline_configuration.execution_mode,
                ExecutionMode.STOP_ON_FAILURE,
            )

        self._snapshot = snapshot
        self._pipeline: Optional["DynamicPipeline"] = None

        worker_count = handle_int_env_var(
            ENV_ZENML_DYNAMIC_PIPELINE_WORKER_COUNT, default=10
        )
        self._executor = ThreadPoolExecutor(max_workers=worker_count)
        if orchestrator:
            self._orchestrator = orchestrator
        else:
            self._orchestrator = Stack.from_model(snapshot.stack).orchestrator

        self._step_operator = Stack.from_model(snapshot.stack).step_operator
        self._futures: Dict[str, "StepFuture"] = {}
        self._invocation_ids: Set[str] = set()

        self._run, self._orchestrator_run_id = self._prepare_run(run)
        self._existing_step_runs = self._run.steps

        self._steps_to_monitor: Dict[str, "StepRunResponse"] = {}
        self._shutdown_event = threading.Event()

    def _prepare_run(
        self, run: Optional["PipelineRunResponse"]
    ) -> Tuple["PipelineRunResponse", str]:
        """Prepare the pipeline run.

        This method either creates a new run if not provided or potentially
        updates the existing run with the orchestrator run ID.

        Args:
            run: The pipeline run to prepare.

        Returns:
            The prepared pipeline run and the orchestrator run ID.
        """
        if run and run.orchestrator_run_id:
            orchestrator_run_id = run.orchestrator_run_id
        else:
            orchestrator_run_id = self._orchestrator.get_orchestrator_run_id()

        if run and not run.orchestrator_run_id:
            run = Client().zen_store.update_run(
                run_id=run.id,
                run_update=PipelineRunUpdate(
                    orchestrator_run_id=orchestrator_run_id,
                ),
            )
        else:
            existing_runs = Client().list_pipeline_runs(
                snapshot_id=self._snapshot.id,
                orchestrator_run_id=orchestrator_run_id,
            )
            if existing_runs.total == 1:
                run = existing_runs.items[0]
            else:
                run = create_placeholder_run(
                    snapshot=self._snapshot,
                    orchestrator_run_id=orchestrator_run_id,
                )

        return run, orchestrator_run_id

    @property
    def pipeline(self) -> "DynamicPipeline":
        """The pipeline that the runner is executing.

        Raises:
            RuntimeError: If the pipeline can't be loaded.

        Returns:
            The pipeline that the runner is executing.
        """
        if self._pipeline is None:
            if (
                not self._snapshot.pipeline_spec
                or not self._snapshot.pipeline_spec.source
            ):
                raise RuntimeError("Missing pipeline source for snapshot.")

            pipeline = source_utils.load(self._snapshot.pipeline_spec.source)
            if not isinstance(pipeline, DynamicPipeline):
                raise RuntimeError(
                    "Invalid pipeline source: "
                    f"{self._snapshot.pipeline_spec.source.import_path}"
                )
            pipeline = copy.deepcopy(pipeline)
            pipeline._configuration = self._snapshot.pipeline_configuration
            self._pipeline = pipeline

        return self._pipeline

    def _monitoring_loop(self) -> None:
        """Monitoring loop.

        This should run in a separate thread and monitors the running steps.
        """
        from zenml.config.step_configurations import Step

        monitoring_interval = handle_float_env_var(
            ENV_ZENML_DYNAMIC_PIPELINE_MONITORING_INTERVAL, default=10.0
        )
        monitoring_delay = handle_float_env_var(
            ENV_ZENML_DYNAMIC_PIPELINE_MONITORING_DELAY, default=0.1
        )

        while not self._shutdown_event.is_set():
            start_time = time.time()
            finished_step_runs = []

            # Copy the steps to avoid the dictionary getting modified by
            # the main thread while we're iterating over it.
            for invocation_id, step_run in list(
                self._steps_to_monitor.items()
            ):
                infra_status: Optional[ExecutionStatus] = None
                try:
                    infra_status = self._get_isolated_step_infra_status(
                        step_run
                    )
                except Exception as e:
                    logger.error(
                        "Failed to get infra status for step `%s`: %s",
                        invocation_id,
                        e,
                    )
                else:
                    logger.debug(
                        "Step `%s` infra status: %s.",
                        invocation_id,
                        infra_status,
                    )

                if infra_status in [
                    ExecutionStatus.INITIALIZING,
                    ExecutionStatus.PROVISIONING,
                    ExecutionStatus.RUNNING,
                ]:
                    # Step still running on the infra side, no need to
                    # do anything here.
                    continue

                step_run = Client().get_run_step(step_run.id)
                # Store the refreshed step run
                self._steps_to_monitor[invocation_id] = step_run

                db_status = step_run.status

                if infra_status is None and db_status in [
                    ExecutionStatus.INITIALIZING,
                    ExecutionStatus.PROVISIONING,
                    ExecutionStatus.RUNNING,
                ]:
                    # Step is running in the DB and we have no information on
                    # the infra side, so we wait for the next
                    # monitoring interval to check again.
                    continue

                finished_step_runs.append(invocation_id)

                if db_status == ExecutionStatus.STOPPING:
                    # The step runner sets the status to STOPPING when receiving
                    # a heartbeat response that the pipeline should be stopped.
                    # We now update this status to STOPPED.
                    step_run = publish_stopped_step_run(step_run.id)
                elif (
                    infra_status
                    in [ExecutionStatus.FAILED, ExecutionStatus.STOPPED]
                    and db_status == ExecutionStatus.RUNNING
                ):
                    # Step failed/stopped on the infra side, but the
                    # code failed before it could report the status back to us.
                    step_run = publish_failed_step_run(step_run.id)

                # Get the updated status that we might have just published
                db_status = step_run.status

                duration_string = (
                    string_utils.get_human_readable_time(
                        step_run.duration.total_seconds()
                    )
                    if step_run.duration
                    else None
                )
                if db_status.is_successful:
                    if duration_string:
                        logger.info(
                            "Step `%s` finished successfully in %s.",
                            invocation_id,
                            duration_string,
                        )
                    else:
                        logger.info(
                            "Step `%s` finished successfully.",
                            invocation_id,
                        )
                elif db_status.is_failed:
                    if duration_string:
                        logger.error(
                            "Step `%s` failed after %s.",
                            invocation_id,
                            duration_string,
                        )
                    else:
                        logger.error(
                            "Step `%s` failed.",
                            invocation_id,
                        )
                elif db_status == ExecutionStatus.STOPPED:
                    if duration_string:
                        logger.info(
                            "Step `%s` stopped after %s.",
                            invocation_id,
                            duration_string,
                        )
                    else:
                        logger.info("Step `%s` stopped.", invocation_id)
                elif db_status == ExecutionStatus.RETRYING:
                    # The server set the status of the step run to retrying ->
                    # The server expects the step to be retried.
                    remaining_retries = get_remaining_retries(
                        step_run=step_run
                    )

                    if duration_string:
                        logger.error(
                            "Step `%s` failed after %s. Remaining retries: %d.",
                            invocation_id,
                            duration_string,
                            remaining_retries,
                        )
                    else:
                        logger.error(
                            "Step `%s` failed. Remaining retries: %d.",
                            invocation_id,
                            remaining_retries,
                        )

                    if remaining_retries > 0:
                        step = Step(
                            spec=step_run.spec,
                            config=step_run.config,
                            step_config_overrides=step_run.config,
                        )
                        self._queue_concurrent_isolated_step(step=step)

                time.sleep(monitoring_delay)

            for invocation_id in finished_step_runs:
                self._steps_to_monitor.pop(invocation_id)

            duration = time.time() - start_time
            time_to_sleep = max(0, monitoring_interval - duration)
            self._shutdown_event.wait(timeout=time_to_sleep)

    def _maybe_stop_isolated_steps(self) -> None:
        """Maybe stop all isolated in progress steps.

        This method will only stop concurrent isolated steps if the execution
        mode is FAIL_FAST.
        """
        if (
            self._snapshot.pipeline_configuration.execution_mode
            != ExecutionMode.FAIL_FAST
        ):
            return

        logger.info("Stopping isolated steps.")

        for step_run in list(self._steps_to_monitor.values()):
            try:
                self._stop_isolated_step(step_run)
            except Exception:
                logger.exception("Failed to stop step `%s`.", step_run.name)

    def _get_isolated_step_infra_status(
        self, step_run: "StepRunResponse"
    ) -> Optional[ExecutionStatus]:
        """Get the status of an isolated step from the component infrastructure.

        Args:
            step_run: The step run to get the status of.

        Returns:
            The status of the step run.
        """
        if step_run.config.step_operator:
            assert self._step_operator
            return self._step_operator.get_status(step_run)
        else:
            return self._orchestrator.get_isolated_step_status(step_run)

    def _stop_isolated_step(self, step_run: "StepRunResponse") -> None:
        """Stop an isolated step.

        Args:
            step_run: The step run to stop.
        """
        if step_run.config.step_operator:
            assert self._step_operator
            self._step_operator.cancel(step_run)
        else:
            self._orchestrator.stop_isolated_step(step_run)

    def run_pipeline(self) -> None:
        """Run the pipeline.

        Raises:
            Exception: If the pipeline run failed.
        """
        logs_context: ContextManager[Any] = nullcontext()
        if is_pipeline_logging_enabled(self._snapshot.pipeline_configuration):
            logs_context = setup_logging_context(
                source="orchestrator", pipeline_run=self._run
            )

        with logs_context:
            if self._run.status.is_finished:
                logger.info("Run `%s` is already finished.", str(self._run.id))
                return
            elif self._run.status in {
                ExecutionStatus.RESUMING,
                ExecutionStatus.PAUSED,
                ExecutionStatus.RETRYING,
            }:
                self._run = Client().zen_store.update_run(
                    run_id=self._run.id,
                    run_update=PipelineRunUpdate(
                        status=ExecutionStatus.RUNNING,
                    ),
                )
                logger.info("Resuming run `%s`.", str(self._run.id))
            elif self._run.status == ExecutionStatus.RUNNING:
                logger.info("Continuing existing run `%s`.", str(self._run.id))
            elif self._run.status in {
                ExecutionStatus.INITIALIZING,
                ExecutionStatus.PROVISIONING,
            }:
                # Set the run status to running already, in case no steps start
                # immediately which would otherwise cause the run to be stuck in
                # some init state.
                self._run = Client().zen_store.update_run(
                    run_id=self._run.id,
                    run_update=PipelineRunUpdate(
                        status=ExecutionStatus.RUNNING,
                    ),
                )

            assert self._snapshot.stack

            with (
                InMemoryArtifactCache(),
                env_utils.temporary_runtime_environment(
                    self._snapshot.pipeline_configuration, self._snapshot.stack
                ),
                DynamicPipelineRunContext(
                    pipeline=self.pipeline,
                    run=self._run,
                    snapshot=self._snapshot,
                    runner=self,
                ),
            ):
                monitoring_thread = threading.Thread(
                    name="DynamicPipelineRunner-Monitoring-Loop",
                    target=lambda: context_utils.run_in_current_context(
                        self._monitoring_loop
                    ),
                    daemon=True,
                )
                monitoring_thread.start()
                if not self._run.triggered_by_deployment:
                    # Only run the init hook if the run is not triggered by
                    # a deployment, as the deployment service will have
                    # already run the init hook.
                    self._orchestrator.run_init_hook(snapshot=self._snapshot)

                try:
                    # TODO: what should be allowed as pipeline returns?
                    #  (artifacts, json serializable, anything?)
                    #  how do we show it in the UI?
                    params = self.pipeline.configuration.parameters or {}
                    self.pipeline._call_entrypoint(**params)
                    # The pipeline function finished successfully, but some
                    # steps might still be running. We now wait for all of
                    # them and raise any exceptions that occurred.
                    self.await_all_step_futures()
                except _WaitConditionPollTimeout:
                    logger.info("Pausing pipeline run `%s`.", self._run.id)
                except _WaitConditionAborted:
                    logger.info(
                        "Stopping pipeline run `%s` because a wait condition "
                        "was aborted.",
                        self._run.id,
                    )
                except Exception as e:
                    exception_info = (
                        exception_utils.collect_exception_information(
                            exception=e,
                            user_func=self.pipeline.entrypoint,
                        )
                    )
                    # TODO: this call already invalidates the token, so
                    # the steps will keep running but won't be able to
                    # report their status back to ZenML.
                    publish_failed_pipeline_run(
                        self._run.id, exception_info=exception_info
                    )
                    raise
                finally:
                    self._shutdown_event.set()

                    if not self._run.triggered_by_deployment:
                        # Only run the cleanup hook if the run is not
                        # triggered by a deployment, as the deployment
                        # service will have already run the cleanup hook.
                        self._orchestrator.run_cleanup_hook(
                            snapshot=self._snapshot
                        )

                    self._maybe_stop_isolated_steps()
                    self._executor.shutdown(wait=True, cancel_futures=True)
                    monitoring_thread.join()

                self._run = Client().zen_store.get_run(
                    self._run.id, hydrate=False
                )
                if self._run.status == ExecutionStatus.RUNNING:
                    publish_successful_pipeline_run(self._run.id)
                    logger.info("Pipeline completed successfully.")

    @overload
    def launch_step(
        self,
        step: "BaseStep",
        id: Optional[str],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
        group: Optional["GroupInfo"] = None,
        concurrent: Literal[False] = False,
    ) -> StepRunOutputs: ...

    @overload
    def launch_step(
        self,
        step: "BaseStep",
        id: Optional[str],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
        group: Optional["GroupInfo"] = None,
        concurrent: Literal[True] = True,
    ) -> "StepFuture": ...

    def launch_step(
        self,
        step: "BaseStep",
        id: Optional[str],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
        group: Optional["GroupInfo"] = None,
        concurrent: bool = False,
    ) -> Union[StepRunOutputs, "StepFuture"]:
        """Launch a step.

        Args:
            step: The step to launch.
            id: The invocation ID of the step.
            args: The arguments for the step function.
            kwargs: The keyword arguments for the step function.
            after: The step run output futures to wait for.
            group: The group information for this step.
            concurrent: Whether to launch the step concurrently.

        # noqa: DAR401
        Raises:
            BaseException: If the step failed.

        Returns:
            The step run outputs or a future for the step run outputs.
        """
        step = step.copy()
        step_config = None
        if self._run and self._run.triggered_by_deployment:
            # Deployment-specific step overrides
            step_config = StepConfigurationUpdate(
                enable_cache=False,
                step_operator=None,
                parameters={},
                runtime=StepRuntime.INLINE,
                group=group,
            )
        elif group:
            step_config = StepConfigurationUpdate(
                group=group,
            )

        inputs = convert_to_keyword_arguments(step.entrypoint, args, kwargs)

        invocation_id = compute_invocation_id(
            existing_invocations=self._invocation_ids,
            step=step,
            custom_id=id,
            allow_suffix=not id,
        )
        self._invocation_ids.add(invocation_id)

        if waiting_for_steps := _get_running_upstream_steps(inputs, after):
            logger.info(
                "Waiting for step(s) `%s` to finish before executing step `%s`.",
                ", ".join(waiting_for_steps),
                invocation_id,
            )

        compiled_step = compile_dynamic_step_invocation(
            snapshot=self._snapshot,
            pipeline=self.pipeline,
            step=step,
            invocation_id=invocation_id,
            inputs=inputs,
            pipeline_docker_settings=self._snapshot.pipeline_configuration.docker_settings,
            after=after,
            config=step_config,
        )

        step_run = self._existing_step_runs.get(invocation_id)
        runtime = get_step_runtime(
            step_config=compiled_step.config,
            pipeline_docker_settings=self._snapshot.pipeline_configuration.docker_settings,
            orchestrator=self._orchestrator,
        )
        remaining_retries = None

        if step_run:
            old_config = step_run.config.model_copy(deep=True)
            # The server includes the date/time substitutions based on the
            # actual DB start time of the pipeline run. We want to compare this
            # to the newly compiled step config, so we remove them.
            old_config.substitutions.pop("date")
            old_config.substitutions.pop("time")

            if (
                old_config.model_dump_json()
                != compiled_step.config.model_dump_json()
            ):
                # We use the old config here to keep the behavior aligned across
                # all step retries: When the orchestration environment is
                # restarted while an isolated step is running, the code below
                # will start monitoring that existing execution. If that step
                # then fails, the monitoring loop will restart it with the
                # configuration of the step run that just failed, which is the
                # old (=compiled in previous orchestration environment) config.
                compiled_step = Step(
                    spec=step_run.spec,
                    config=old_config,
                    step_config_overrides=old_config,
                )
                logger.warning(
                    "Configuration for step `%s` changed since the the "
                    "orchestration environment was restarted. If the step "
                    "needs to be retried, it will use the old configuration.",
                    step_run.name,
                )

            if step_run.status.is_successful:
                # The step finished successfully, but we still need to return
                # a future in case the step was launched concurrently so the
                # caller gets the correct object back.
                if concurrent:
                    future = StepFuture(
                        wrapped=_IsolatedStepFuture(
                            pipeline_run_id=self._run.id,
                            invocation_id=invocation_id,
                        ),
                        output_keys=list(compiled_step.config.outputs),
                    )
                    self._futures[invocation_id] = future
                    return future
                else:
                    return load_step_run_outputs(step_run.id)

            if (
                runtime == StepRuntime.INLINE
                and step_run.status == ExecutionStatus.RUNNING
            ):
                # Inline steps that are in running state didn't have the
                # chance to report their failure back to ZenML before the
                # orchestration environment was shut down. But there is no
                # way that they're actually still running if we're in a new
                # orchestration environment, so we mark them as failed and
                # potentially restart them depending on the retry config.
                step_run = publish_failed_step_run(step_run.id)

            if step_run.status.is_failed:
                # If the step is running concurrently, we only raise the
                # exception once the future is awaited.
                if concurrent:
                    future = StepFuture(
                        wrapped=_IsolatedStepFuture(
                            pipeline_run_id=self._run.id,
                            invocation_id=invocation_id,
                        ),
                        output_keys=list(compiled_step.config.outputs),
                    )
                    self._futures[invocation_id] = future
                    return future
                else:
                    raise exception_utils.reconstruct_exception(
                        exception_info=step_run.exception_info,
                        fallback_message=(
                            f"Step `{invocation_id}` failed with status "
                            f"`{step_run.status}`."
                        ),
                    )

            remaining_retries = get_remaining_retries(step_run=step_run)

            if step_run.status == ExecutionStatus.RUNNING:
                logger.info(
                    "Restarting the monitoring of existing step `%s` "
                    "(ID: %s). Remaining retries: %d",
                    step_run.name,
                    step_run.id,
                    remaining_retries,
                )
                self._steps_to_monitor[invocation_id] = step_run
                monitoring_future = StepFuture(
                    wrapped=_IsolatedStepFuture(
                        pipeline_run_id=self._run.id,
                        invocation_id=invocation_id,
                    ),
                    output_keys=list(compiled_step.config.outputs),
                )
                self._futures[invocation_id] = monitoring_future
                return monitoring_future

        if not concurrent:
            step_run = self._run_sync_step(
                step=compiled_step, remaining_retries=remaining_retries
            )
            return load_step_run_outputs(step_run.id)
        elif runtime == StepRuntime.INLINE:
            return self._queue_concurrent_inline_step(
                step=compiled_step, remaining_retries=remaining_retries
            )
        else:
            return self._queue_concurrent_isolated_step(step=compiled_step)

    def _queue_concurrent_isolated_step(self, step: "Step") -> StepFuture:
        """Queue a concurrent isolated step.

        Args:
            step: The step to queue.

        Returns:
            The step future.
        """

        def _launch_no_wait() -> StepRunResponse:
            step_run = launch_step(
                snapshot=self._snapshot,
                step=step,
                orchestrator_run_id=self._orchestrator_run_id,
                # The monitoring loop is responsible for monitoring and retrying
                # steps.
                wait=False,
                retry=False,
            )
            self._steps_to_monitor[step.spec.invocation_id] = step_run
            return step_run

        concurrent_future = self._executor.submit(
            context_utils.run_in_current_context, _launch_no_wait
        )
        future = StepFuture(
            wrapped=_IsolatedStepFuture(
                wrapped=concurrent_future,
                pipeline_run_id=self._run.id,
                invocation_id=step.spec.invocation_id,
            ),
            output_keys=list(step.config.outputs),
        )
        self._futures[step.spec.invocation_id] = future
        return future

    def _queue_concurrent_inline_step(
        self, step: "Step", remaining_retries: Optional[int] = None
    ) -> StepFuture:
        """Queue a concurrent inline step.

        Args:
            step: The step to queue.
            remaining_retries: The remaining retries for the step.

        Returns:
            The step future.
        """

        def _launch_and_wait() -> StepRunResponse:
            return self._run_sync_step(
                step=step, remaining_retries=remaining_retries
            )

        concurrent_future = self._executor.submit(
            context_utils.run_in_current_context, _launch_and_wait
        )
        future = StepFuture(
            wrapped=_InlineStepFuture(
                wrapped=concurrent_future,
                invocation_id=step.spec.invocation_id,
            ),
            output_keys=list(step.config.outputs),
        )
        self._futures[step.spec.invocation_id] = future
        return future

    def _run_sync_step(
        self,
        step: "Step",
        remaining_retries: Optional[int] = None,
    ) -> StepRunResponse:
        """Run a synchronous step.

        Args:
            step: The step to run.
            remaining_retries: The remaining retries for the step.

        Returns:
            The step run response.
        """
        return launch_step(
            snapshot=self._snapshot,
            step=step,
            orchestrator_run_id=self._orchestrator_run_id,
            wait=True,
            retry=True,
            remaining_retries=remaining_retries,
        )

    def map(
        self,
        step: "BaseStep",
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
        product: bool = False,
    ) -> "MapResultsFuture":
        """Map over step inputs.

        Args:
            step: The step to run.
            args: The arguments for the step function.
            kwargs: The keyword arguments for the step function.
            after: The step run output futures to wait for before executing the
                steps.
            product: Whether to produce a cartesian product of the mapped
                inputs.

        Returns:
            A future that represents the map results.
        """
        kwargs = convert_to_keyword_arguments(step.entrypoint, args, kwargs)
        kwargs = await_step_inputs(kwargs)
        step_inputs = expand_mapped_inputs(kwargs, product=product)

        # This will overwrite any user-configured groups for the step, but
        # capturing the mapping information is more important until we introduce
        # a more flexible group system.
        group_info = GroupInfo(
            id=str(uuid4()),
            name=step.name,
            type=GroupType.MAP,
        )

        step_futures = [
            self.launch_step(
                step,
                id=None,
                args=(),
                kwargs=inputs,
                after=after,
                group=group_info,
                concurrent=True,
            )
            for inputs in step_inputs
        ]

        return MapResultsFuture(futures=step_futures)

    @overload
    def wait(
        self,
        schema: Type[T],
        type: RunWaitConditionType = RunWaitConditionType.EXTERNAL_INPUT,
        timeout: int = 600,
        poll_interval: int = 5,
        question: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
    ) -> T: ...

    @overload
    def wait(
        self,
        schema: object = None,
        type: RunWaitConditionType = RunWaitConditionType.EXTERNAL_INPUT,
        timeout: int = 600,
        poll_interval: int = 5,
        question: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
    ) -> Any: ...

    def wait(
        self,
        schema: Optional[Any] = None,
        type: RunWaitConditionType = RunWaitConditionType.EXTERNAL_INPUT,
        timeout: int = 600,
        poll_interval: int = 5,
        question: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
    ) -> Any:
        """Create and poll a run wait condition.

        Args:
            schema: Optional expected output type for the resolved result.
            type: Wait condition type.
            timeout: Maximum time in seconds to poll before pausing.
            poll_interval: Poll interval in seconds.
            question: Optional question shown to external actors.
            metadata: Optional metadata attached to the condition.
            name: Optional deterministic wait condition name.
            after: Optional upstream futures that must finish before waiting.

        Raises:
            RuntimeError: If called outside the dynamic pipeline function.
            _WaitConditionAborted: If the wait condition was aborted.
            _WaitConditionPollTimeout: If the wait condition polling timed out.
            KeyboardInterrupt: If interrupted while waiting.
            BaseException: If polling fails after the lease is abandoned.

        Returns:
            The resolved wait condition value.
        """
        from zenml.execution.pipeline.dynamic.run_context import (
            DynamicPipelineRunContext,
        )
        from zenml.steps.step_context import StepContext
        from zenml.utils.time_utils import utc_now

        context = DynamicPipelineRunContext.get()
        if not context:
            raise RuntimeError(
                "`zenml.wait(...)` can only be used inside dynamic pipelines."
            )

        if StepContext.is_active():
            raise RuntimeError(
                "`zenml.wait(...)` cannot be called inside a step function. "
                "Use it only in the pipeline function."
            )

        wait_condition_name = name or context.next_wait_condition_name()

        if isinstance(after, BaseStepFuture):
            after.result()
        elif isinstance(after, MapResultsFuture):
            for future in after:
                future.result()
        elif isinstance(after, Sequence):
            for item in after:
                if isinstance(item, BaseStepFuture):
                    item.result()
                elif isinstance(item, MapResultsFuture):
                    for future in item:
                        future.result()

        condition = Client().zen_store.create_run_wait_condition(
            RunWaitConditionRequest(
                project=self._run.project_id,
                run=self._run.id,
                name=wait_condition_name,
                type=type,
                question=question,
                metadata=metadata or {},
                data_schema=pydantic_utils.get_json_schema_for_type(schema)
                if schema
                else None,
            )
        )
        state = self._handle_wait_condition_state(
            condition=condition, schema=schema
        )
        if state.is_terminal:
            return state.value

        if poll_interval <= 0:
            logger.debug(
                "Non-positive poll interval provided, falling back to 5 seconds."
            )
            poll_interval = 5

        logger.info(
            "Waiting on wait condition `%s` (type=%s, timeout=%ss, poll=%ss).",
            wait_condition_name,
            type.value,
            timeout,
            poll_interval,
        )
        deadline = time.time() + timeout

        try:
            with maybe_enable_interactive_wait_prompt(
                orchestrator=self._orchestrator,
                condition=condition,
            ) as interactive_prompting_enabled:
                # We keep polling until the deadline is reached and all ongoing
                # steps have finished.
                while time.time() < deadline or self.has_in_progress_steps():
                    lease_now = utc_now()
                    status = (
                        Client().zen_store.update_run_wait_condition_lease(
                            run_wait_condition_id=condition.id,
                            lease_update=RunWaitConditionLeaseUpdate(
                                poller_instance_id=self._orchestrator_run_id,
                                poller_lease_expires_at=lease_now
                                + timedelta(
                                    seconds=max(15, poll_interval * 2)
                                ),
                            ),
                        )
                    )
                    if status != RunWaitConditionStatus.PENDING:
                        condition = Client().zen_store.get_run_wait_condition(
                            condition.id, hydrate=True
                        )
                    state = self._handle_wait_condition_state(
                        condition=condition, schema=schema
                    )
                    if state.is_terminal:
                        return state.value

                    if interactive_prompting_enabled:
                        # If we're running interactively, we sleep until the
                        # polling interval is reached or the user submitted
                        # input in their terminal.
                        poll_interactive_wait_condition_input(
                            condition=condition,
                            poll_interval=poll_interval,
                        )
                    else:
                        time.sleep(poll_interval)
        except _WaitConditionAborted:
            raise
        except KeyboardInterrupt:
            try:
                Client().zen_store.resolve_run_wait_condition(
                    run_wait_condition_id=condition.id,
                    resolve_request=RunWaitConditionResolveRequest(
                        resolution=RunWaitConditionResolution.ABORT,
                    ),
                )
            except Exception as e:
                logger.warning(
                    "Failed to abort wait condition `%s` after keyboard "
                    "interrupt: %s",
                    condition.id,
                    e,
                )
            raise
        except BaseException:
            try:
                Client().zen_store.update_run_wait_condition_lease(
                    run_wait_condition_id=condition.id,
                    lease_update=RunWaitConditionLeaseUpdate(
                        poller_instance_id=self._orchestrator_run_id,
                        poller_lease_expires_at=utc_now(),
                        mode=RunWaitConditionLeaseMode.ABANDON,
                    ),
                )
            except Exception as e:
                logger.warning(
                    "Failed to abandon wait condition `%s` after wait loop "
                    "failure: %s",
                    condition.id,
                    e,
                )
            raise

        status = Client().zen_store.update_run_wait_condition_lease(
            run_wait_condition_id=condition.id,
            lease_update=RunWaitConditionLeaseUpdate(
                poller_instance_id=self._orchestrator_run_id,
                poller_lease_expires_at=utc_now(),
                mode=RunWaitConditionLeaseMode.FINALIZE,
            ),
        )
        if status != RunWaitConditionStatus.PENDING:
            condition = Client().zen_store.get_run_wait_condition(
                condition.id, hydrate=True
            )
        state = self._handle_wait_condition_state(
            condition=condition, schema=schema
        )
        if state.is_terminal:
            return state.value

        raise _WaitConditionPollTimeout(
            f"Wait condition `{condition.name}` polling timed out."
        )

    @staticmethod
    def _handle_wait_condition_state(
        condition: "RunWaitConditionResponse",
        schema: Optional[Any] = None,
    ) -> _WaitConditionState:
        """Handle wait conditions.

        Args:
            condition: The latest wait condition state.
            schema: Optional schema used to parse the resolved result.

        Raises:
            _WaitConditionAborted: If the condition was resolved with abort.

        Returns:
            State indicating whether the condition is terminal and its value.
        """
        if condition.status == RunWaitConditionStatus.PENDING:
            return _WaitConditionState(is_terminal=False, value=None)

        if condition.resolution == RunWaitConditionResolution.ABORT:
            raise _WaitConditionAborted(
                f"Wait condition `{condition.name}` resolved with abort."
            )

        if schema is None:
            return _WaitConditionState(is_terminal=True, value=None)

        return _WaitConditionState(
            is_terminal=True,
            value=TypeAdapter(schema).validate_python(condition.result),
        )

    def await_all_step_futures(self) -> None:
        """Await all step futures."""
        for future in list(self._futures.values()):
            future.wait()
        self._futures = {}

    def has_in_progress_steps(self) -> bool:
        """Check if there are any in-progress steps.

        Returns:
            True if there are any in-progress steps, False otherwise.
        """
        return any(future.running() for future in list(self._futures.values()))


def compile_dynamic_step_invocation(
    snapshot: "PipelineSnapshotResponse",
    pipeline: "DynamicPipeline",
    step: "BaseStep",
    invocation_id: str,
    inputs: Dict[str, Any],
    pipeline_docker_settings: "DockerSettings",
    after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
    config: Optional[StepConfigurationUpdate] = None,
) -> "Step":
    """Compile a dynamic step invocation.

    Args:
        snapshot: The snapshot.
        pipeline: The dynamic pipeline.
        step: The step to compile.
        invocation_id: The invocation ID of the step.
        inputs: The inputs for the step function.
        pipeline_docker_settings: The Docker settings of the parent pipeline.
        after: The step run output futures to wait for.
        config: The configuration for the step.

    Returns:
        The compiled step.
    """
    upstream_steps = set()

    if isinstance(after, BaseStepFuture):
        after.result()
        upstream_steps.add(after.invocation_id)
    elif isinstance(after, MapResultsFuture):
        for future in after:
            future.result()
            upstream_steps.add(future.invocation_id)
    elif isinstance(after, Sequence):
        for item in after:
            if isinstance(item, BaseStepFuture):
                item.result()
                upstream_steps.add(item.invocation_id)
            elif isinstance(item, MapResultsFuture):
                for future in item:
                    future.result()
                    upstream_steps.add(future.invocation_id)

    inputs = await_step_inputs(inputs)

    for value in inputs.values():
        if isinstance(value, OutputArtifact):
            upstream_steps.add(value.step_name)

        if (
            isinstance(value, Sequence)
            and value
            and all(isinstance(item, OutputArtifact) for item in value)
        ):
            upstream_steps.update(item.step_name for item in value)

    input_artifacts = {}
    external_artifacts = {}
    for name, value in inputs.items():
        if isinstance(value, OutputArtifact):
            input_artifacts[name] = [
                StepArtifact(
                    invocation_id=value.step_name,
                    output_name=value.output_name,
                    annotation=OutputSignature(resolved_annotation=Any),
                    pipeline=pipeline,
                    chunk_index=value.chunk_index,
                    chunk_size=value.chunk_size,
                )
            ]
        elif (
            isinstance(value, list)
            and value
            and all(isinstance(item, OutputArtifact) for item in value)
        ):
            input_artifacts[name] = [
                StepArtifact(
                    invocation_id=item.step_name,
                    output_name=item.output_name,
                    annotation=OutputSignature(resolved_annotation=Any),
                    pipeline=pipeline,
                    chunk_index=item.chunk_index,
                    chunk_size=item.chunk_size,
                )
                for item in value
            ]
        elif isinstance(value, (ArtifactVersionResponse, ExternalArtifact)):
            external_artifacts[name] = value
        else:
            # TODO: should some of these be parameters?
            external_artifacts[name] = ExternalArtifact(value=value)

    if template := get_config_template(snapshot, step, pipeline):
        logger.debug(
            "Using config template `%s` for step `%s`",
            template.spec.invocation_id,
            invocation_id,
        )
        step._configuration = template.config.model_copy(
            update={"template": template.spec.invocation_id}
        )

    default_parameters = {
        key: value
        for key, value in convert_to_keyword_arguments(
            step.entrypoint, (), inputs, apply_defaults=True
        ).items()
        if key not in inputs and key not in step.configuration.parameters
    }

    step_invocation = StepInvocation(
        id=invocation_id,
        step=step,
        input_artifacts=input_artifacts,
        external_artifacts=external_artifacts,
        default_parameters=default_parameters,
        upstream_steps=upstream_steps,
        pipeline=pipeline,
        model_artifacts_or_metadata={},
        client_lazy_loaders={},
        parameters={},
    )

    compiled_step = Compiler()._compile_step_invocation(
        invocation=step_invocation,
        stack=Client().active_stack,
        step_config=config,
        pipeline=pipeline,
    )

    if not compiled_step.config.docker_settings.skip_build:
        if template:
            if (
                template.config.docker_settings
                != compiled_step.config.docker_settings
            ):
                logger.warning(
                    "Custom Docker settings specified for step %s will be "
                    "ignored. The image built for template %s will be used "
                    "instead.",
                    invocation_id,
                    template.spec.invocation_id,
                )
        elif compiled_step.config.docker_settings != pipeline_docker_settings:
            logger.warning(
                "Custom Docker settings specified for step %s will be "
                "ignored. The image built for the pipeline will be used "
                "instead.",
                invocation_id,
            )

    return compiled_step


def get_step_runtime(
    step_config: "StepConfiguration",
    pipeline_docker_settings: "DockerSettings",
    orchestrator: Optional["BaseOrchestrator"] = None,
) -> StepRuntime:
    """Determine if a step should be run in process.

    Args:
        step_config: The step configuration.
        pipeline_docker_settings: The Docker settings of the parent pipeline.
        orchestrator: The orchestrator to use. If not provided, the
            orchestrator will be inferred from the active stack.

    Returns:
        The runtime for the step.
    """
    if step_config.step_operator:
        return StepRuntime.ISOLATED

    if not orchestrator:
        orchestrator = Client().active_stack.orchestrator

    if not orchestrator.can_run_isolated_steps:
        return StepRuntime.INLINE

    runtime = step_config.runtime

    if runtime is None:
        if not step_config.resource_settings.empty:
            runtime = StepRuntime.ISOLATED
        elif step_config.docker_settings != pipeline_docker_settings:
            runtime = StepRuntime.ISOLATED
        else:
            runtime = StepRuntime.INLINE

    return runtime


def get_config_template(
    snapshot: "PipelineSnapshotResponse",
    step: "BaseStep",
    pipeline: "DynamicPipeline",
) -> Optional["Step"]:
    """Get the config template for a step executed in a dynamic pipeline.

    Args:
        snapshot: The snapshot of the pipeline.
        step: The step to get the config template for.
        pipeline: The dynamic pipeline that the step is being executed in.

    Returns:
        The config template for the step.
    """
    for index, step_ in enumerate(pipeline.depends_on):
        if step_._static_id == step._static_id:
            break
    else:
        return None

    return list(snapshot.step_configurations.values())[index]


def expand_mapped_inputs(
    inputs: Dict[str, Any],
    product: bool = False,
) -> List[Dict[str, Any]]:
    """Find the mapped and unmapped inputs of a step.

    Args:
        inputs: The step function inputs.
        product: Whether to produce a cartesian product of the mapped inputs.

    Raises:
        RuntimeError: If no mapped inputs are found or the input combinations
            are not valid.

    Returns:
        The step inputs.
    """
    static_inputs: Dict[str, Any] = {}
    mapped_input_names: List[str] = []
    mapped_inputs: List[Tuple[OutputArtifact, ...]] = []

    for key, value in inputs.items():
        if isinstance(value, _Unmapped):
            static_inputs[key] = value.value
        elif isinstance(value, OutputArtifact):
            if value.item_count is None:
                static_inputs[key] = value
            elif value.item_count == 0:
                raise RuntimeError(
                    f"Artifact `{value.id}` has 0 items and cannot be mapped "
                    "over. Wrap it with the `unmapped(...)` function to pass "
                    "the artifact without mapping over it."
                )
            else:
                mapped_input_names.append(key)
                mapped_inputs.append(
                    tuple(
                        value.chunk(index=i) for i in range(value.item_count)
                    )
                )
        elif (
            isinstance(value, ArtifactVersionResponse)
            and value.item_count is not None
        ):
            static_inputs[key] = value
            logger.warning(
                "Received sequence-like artifact for step input `%s`. Mapping "
                "over artifacts that are not step output artifacts is "
                "currently not supported, and the complete artifact will be "
                "passed to all steps. If you want to silence this warning, "
                "wrap your input with the `unmapped(...)` function.",
                key,
            )
        elif (
            isinstance(value, Sequence)
            and value
            and all(isinstance(item, OutputArtifact) for item in value)
        ):
            # List of step output artifacts, in this case the mapping is over
            # the items of the list
            mapped_input_names.append(key)
            mapped_inputs.append(tuple(value))
        elif isinstance(value, Sequence):
            logger.warning(
                "Received sequence-like data for step input `%s`. Mapping over "
                "data that is not a step output artifact is currently not "
                "supported, and the complete data will be passed to all steps. "
                "If you want to silence this warning, wrap your input with the "
                "`unmapped(...)` function.",
                key,
            )
            static_inputs[key] = value
        else:
            static_inputs[key] = value

    if len(mapped_inputs) == 0:
        raise RuntimeError(
            "No inputs to map over found. When calling `.map(...)` or "
            "`.product(...)` on a step, you need to pass at least one "
            "sequence-like step output of a previous step as input."
        )

    step_inputs = []

    if product:
        for input_combination in itertools.product(*mapped_inputs):
            all_inputs = copy.deepcopy(static_inputs)
            for name, value in zip(mapped_input_names, input_combination):
                all_inputs[name] = value
            step_inputs.append(all_inputs)
    else:
        item_counts = [len(inputs) for inputs in mapped_inputs]
        if not all(count == item_counts[0] for count in item_counts):
            raise RuntimeError(
                f"All mapped input artifacts must have the same "
                "item counts, but you passed artifacts with item counts "
                f"{item_counts}. If you want "
                "to pass sequence-like artifacts without mapping over "
                "them, wrap them with the `unmapped(...)` function."
            )

        for i in range(item_counts[0]):
            all_inputs = copy.deepcopy(static_inputs)
            for name, artifact in zip(
                mapped_input_names,
                [artifact_list[i] for artifact_list in mapped_inputs],
            ):
                all_inputs[name] = artifact
            step_inputs.append(all_inputs)

    return step_inputs


def convert_to_keyword_arguments(
    func: Callable[..., Any],
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    apply_defaults: bool = False,
) -> Dict[str, Any]:
    """Convert function arguments to keyword arguments.

    Args:
        func: The function to convert the arguments to keyword arguments for.
        args: The arguments to convert to keyword arguments.
        kwargs: The keyword arguments to convert to keyword arguments.
        apply_defaults: Whether to apply the function default values.

    Returns:
        The keyword arguments.
    """
    signature = inspect.signature(func, follow_wrapped=True)
    bound_args = signature.bind_partial(*args, **kwargs)
    if apply_defaults:
        bound_args.apply_defaults()

    return bound_args.arguments


def await_step_inputs(inputs: Dict[str, Any]) -> Dict[str, Any]:
    """Await the inputs of a step.

    Args:
        inputs: The inputs of the step.

    Raises:
        RuntimeError: If a step run future with multiple output artifacts is
            passed as an input.

    Returns:
        The awaited inputs.
    """
    result = {}
    for key, value in inputs.items():
        if isinstance(value, MapResultsFuture):
            value = value.futures

        if (
            isinstance(value, Sequence)
            and value
            and all(isinstance(item, StepFuture) for item in value)
        ):
            if any(len(item._output_keys) != 1 for item in value):
                raise RuntimeError(
                    f"Invalid step input `{key}`: Passing a future that refers "
                    "to multiple output artifacts as an input to another step "
                    "is not allowed."
                )
            value = [item.artifacts() for item in value]
        elif isinstance(value, StepFuture):
            if len(value._output_keys) != 1:
                raise RuntimeError(
                    f"Invalid step input `{key}`: Passing a future that refers "
                    "to multiple output artifacts as an input to another step "
                    "is not allowed."
                )
            value = value.artifacts()

        if (
            isinstance(value, Sequence)
            and value
            and all(isinstance(item, ArtifactFuture) for item in value)
        ):
            value = [item.result() for item in value]

        if isinstance(value, ArtifactFuture):
            value = value.result()

        result[key] = value

    return result


def _get_running_upstream_steps(
    inputs: Dict[str, Any],
    after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None],
) -> List[str]:
    """Get all running upstream steps for a step.

    Args:
        inputs: The inputs of the step.
        after: The step run futures to wait for.

    Raises:
        TypeError: If an unexpected future type is passed.

    Returns:
        The list of running upstream steps.
    """
    futures: List[BaseFuture] = []

    for value in inputs.values():
        if isinstance(value, BaseFuture):
            futures.append(value)
        elif isinstance(value, Sequence) and all(
            isinstance(item, BaseFuture) for item in value
        ):
            futures.extend(value)

    if isinstance(after, BaseFuture):
        futures.append(after)
    elif isinstance(after, Sequence):
        futures.extend(after)

    steps = []

    for future in futures:
        if isinstance(future, MapResultsFuture):
            for item in future.futures:
                if item.running():
                    steps.append(item.invocation_id)
        elif isinstance(future, BaseStepFuture):
            if future.running():
                steps.append(future.invocation_id)
        else:
            raise TypeError(f"Unexpected future type: {type(future)}")

    return steps


def get_remaining_retries(step_run: "StepRunResponse") -> int:
    """Get the remaining retries for a step run.

    Args:
        step_run: The step run to get the remaining retries for.

    Returns:
        The remaining retries for the step run.
    """
    max_retries = (
        step_run.config.retry.max_retries if step_run.config.retry else 0
    )
    return max(0, 1 + max_retries - step_run.version)



"""remove secrets manager [1041bc644e0d].

Revision ID: 1041bc644e0d
Revises: 4e1972485075
Create Date: 2023-12-18 18:40:26.612845

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1041bc644e0d"
down_revision = "4e1972485075"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    conn = op.get_bind()
    meta = sa.MetaData()
    meta.reflect(
        only=("stack_component", "stack_composition"), bind=op.get_bind()
    )
    components = sa.Table("stack_component", meta)
    compositions = sa.Table("stack_composition", meta)

    # Find all secrets manager components
    secrets_manager_components = conn.execute(
        sa.select(components.c.id).where(
            components.c.type == "secrets_manager"
        )
    ).all()

    secrets_manager_ids = [i[0] for i in secrets_manager_components]

    # Remove all secrets manager compositions
    conn.execute(
        compositions.delete().where(
            compositions.c.component_id.in_(secrets_manager_ids)
        )
    )

    # Remove all secrets manager components
    conn.execute(
        components.delete().where(components.c.id.in_(secrets_manager_ids))
    )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for triggers."""

from uuid import UUID

from fastapi import APIRouter, Depends, Security

from zenml.constants import (
    API,
    PIPELINE_SNAPSHOTS,
    SCHEDULE_FEATURE,
    TRIGGERS,
    VERSION_1,
)
from zenml.models import (
    TRIGGER_CREATE_TYPE_UNION,
    TRIGGER_RETURN_TYPE_UNION,
    TRIGGER_UPDATE_TYPE_UNION,
    Page,
    TriggerFilter,
)
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.feature_gate.endpoint_utils import check_entitlement
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_create_entity,
    verify_permissions_and_delete_entity,
    verify_permissions_and_get_entity,
    verify_permissions_and_list_entities,
    verify_permissions_and_update_entity,
)
from zenml.zen_server.rbac.models import Action, ResourceType
from zenml.zen_server.rbac.utils import (
    verify_permission,
    verify_permission_for_model,
)
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1 + TRIGGERS,
    tags=["triggers"],
    responses={401: error_response, 403: error_response},
)


@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def create_trigger(
    trigger: TRIGGER_CREATE_TYPE_UNION,
    _: AuthContext = Security(authorize),
) -> TRIGGER_RETURN_TYPE_UNION:
    """Creates a trigger.

    Args:
        trigger: The trigger to create.

    Returns:
        The created trigger.
    """
    check_entitlement(feature=SCHEDULE_FEATURE)

    return verify_permissions_and_create_entity(
        request_model=trigger,
        create_method=zen_store().create_trigger,
    )


@router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def list_triggers(
    trigger_filter_model: TriggerFilter = Depends(
        make_dependable(TriggerFilter)
    ),
    hydrate: bool = False,
    _: AuthContext = Security(authorize),
) -> Page[TRIGGER_RETURN_TYPE_UNION]:
    """Gets a list of triggers.

    Args:
        trigger_filter_model: Filter model used for pagination, sorting,
            filtering
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        List of trigger objects.
    """
    return verify_permissions_and_list_entities(
        filter_model=trigger_filter_model,
        resource_type=ResourceType.TRIGGER,
        list_method=zen_store().list_triggers,
        hydrate=hydrate,
    )


@router.get(
    "/{trigger_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_trigger(
    trigger_id: UUID,
    hydrate: bool = True,
    _: AuthContext = Security(authorize),
) -> TRIGGER_RETURN_TYPE_UNION:
    """Gets a specific trigger using its unique id.

    Args:
        trigger_id: ID of the trigger to retrieve.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        A specific trigger object.
    """
    return verify_permissions_and_get_entity(
        id=trigger_id,
        get_method=zen_store().get_trigger,
        hydrate=hydrate,
    )


@router.put(
    "/{trigger_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def update_trigger(
    trigger_id: UUID,
    trigger_update: TRIGGER_UPDATE_TYPE_UNION,
    _: AuthContext = Security(authorize),
) -> TRIGGER_RETURN_TYPE_UNION:
    """Updates the attributes on a specific trigger using its unique id.

    Args:
        trigger_id: ID of the trigger to update.
        trigger_update: the model containing the attributes to update.

    Returns:
        The updated trigger object.
    """
    check_entitlement(feature=SCHEDULE_FEATURE)

    return verify_permissions_and_update_entity(
        id=trigger_id,
        update_model=trigger_update,
        get_method=zen_store().get_trigger,
        update_method=zen_store().update_trigger,
    )


@router.delete(
    "/{trigger_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def delete_trigger(
    trigger_id: UUID,
    soft: bool,
    _: AuthContext = Security(authorize),
) -> None:
    """Deletes a specific trigger using its unique id.

    Args:
        trigger_id: ID of the trigger to delete.
        soft: Soft deletion will archive the trigger.
    """
    verify_permissions_and_delete_entity(
        id=trigger_id,
        get_method=zen_store().get_trigger,
        delete_method=zen_store().delete_trigger,
        soft=soft,
    )


@router.put(
    "/{trigger_id}" + PIPELINE_SNAPSHOTS + "/{snapshot_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def attach_trigger_to_snapshot(
    trigger_id: UUID,
    snapshot_id: UUID,
    _: AuthContext = Security(authorize),
) -> None:
    """Attaches a trigger to a snapshot.

    Args:
        trigger_id: The ID of the trigger.
        snapshot_id: The ID of the snapshot.
    """
    verify_permission_for_model(
        model=zen_store().get_trigger(trigger_id=trigger_id, hydrate=True),
        action=Action.UPDATE,
    )

    snapshot = zen_store().get_snapshot(snapshot_id=snapshot_id, hydrate=True)

    verify_permission_for_model(
        model=snapshot,
        action=Action.READ,
    )

    verify_permission(
        resource_type=ResourceType.PIPELINE_RUN,
        action=Action.CREATE,
        project_id=snapshot.project_id,
    )

    check_entitlement(feature=SCHEDULE_FEATURE)

    zen_store().attach_trigger_to_snapshot(
        trigger_id=trigger_id,
        snapshot_id=snapshot_id,
    )


@router.delete(
    "/{trigger_id}" + PIPELINE_SNAPSHOTS + "/{snapshot_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def detach_trigger_from_snapshot(
    trigger_id: UUID,
    snapshot_id: UUID,
    _: AuthContext = Security(authorize),
) -> None:
    """Detaches a trigger from a snapshot.

    Args:
        trigger_id: The ID of the trigger.
        snapshot_id: The ID of the snapshot.
    """
    verify_permission_for_model(
        model=zen_store().get_trigger(trigger_id=trigger_id, hydrate=True),
        action=Action.UPDATE,
    )

    verify_permission_for_model(
        model=zen_store().get_snapshot(snapshot_id=snapshot_id, hydrate=True),
        action=Action.READ,
    )

    zen_store().detach_trigger_from_snapshot(
        trigger_id=trigger_id,
        snapshot_id=snapshot_id,
    )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Exception utilities."""

import inspect
import os
import re
import textwrap
import traceback
from typing import Any, Callable, Optional, Type

from zenml.constants import MEDIUMTEXT_MAX_LENGTH
from zenml.logger import get_logger
from zenml.models import (
    ExceptionInfo,
)
from zenml.utils import source_utils

logger = get_logger(__name__)


def collect_exception_information(
    exception: BaseException, user_func: Optional[Callable[..., Any]] = None
) -> ExceptionInfo:
    """Collects the exception information.

    Args:
        exception: The exception to collect information from.
        user_func: The user function that was called.

    Returns:
        The exception information.
    """
    tb = traceback.format_tb(exception.__traceback__)
    line_number = None
    start_index = None

    if user_func and (source_file := inspect.getsourcefile(user_func)):
        try:
            source_file = os.path.abspath(source_file)

            lines, start_line = inspect.getsourcelines(user_func)
            end_line = start_line + len(lines)

            line_pattern = re.compile(
                rf'File "{re.escape(source_file)}", line (\d+),'
            )

            for index, line in enumerate(tb):
                match = line_pattern.search(line)
                if match:
                    potential_line_number = int(match.group(1))
                    if (
                        potential_line_number >= start_line
                        and potential_line_number <= end_line
                    ):
                        line_number = potential_line_number - start_line
                        start_index = index
                        break
        except Exception as e:
            logger.debug("Failed to detect code line: %s", e)

    if start_index is not None:
        # If the code failed while executing user code, we remove the initial
        # part of the traceback that is happening in the ZenML code.
        tb = tb[start_index:]

    tb_bytes = textwrap.dedent("\n".join(tb)).encode()
    tb_bytes = tb_bytes[:MEDIUMTEXT_MAX_LENGTH]

    source = None
    try:
        source = source_utils.resolve(type(exception)).import_path
    except Exception as e:
        logger.debug("Failed to resolve exception source: %s", e)

    message = str(exception) if str(exception) else None

    return ExceptionInfo(
        # Ignore errors when decoding in case we cut off in the middle of an
        # encoded character.
        traceback=tb_bytes.decode(errors="ignore"),
        source=source,
        message=message,
        user_code_line=line_number,
    )


def reconstruct_exception(
    exception_info: Optional[ExceptionInfo], fallback_message: str
) -> BaseException:
    """Reconstruct an exception.

    Args:
        exception_info: Exception information.
        fallback_message: Message to use if the exception cannot be
            reconstructed.

    Returns:
        The reconstructed exception if possible, otherwise a RuntimeError.
    """
    if not exception_info:
        return RuntimeError(fallback_message)

    message = exception_info.message or fallback_message
    source = exception_info.source
    if not source:
        return RuntimeError(message)

    try:
        exception_class: Type[BaseException] = (
            source_utils.load_and_validate_class(
                source=source, expected_class=BaseException
            )
        )
    except Exception as e:
        logger.warning("Failed to load exception source `%s`: %s", source, e)
        return RuntimeError(message)

    try:
        return exception_class(message)
    except Exception as e:
        logger.warning(
            "Failed to instantiate exception `%s` with message `%s`: %s",
            source,
            message,
            e,
        )
        try:
            return exception_class()
        except Exception as e:
            logger.warning(
                "Failed to instantiate exception `%s` without args: %s",
                source,
                e,
            )
            return RuntimeError(message)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQL Model Implementations for Secrets."""

import base64
import json
from typing import Any, Dict, Optional, Sequence, cast
from uuid import UUID

from sqlalchemy import TEXT, VARCHAR, Column, UniqueConstraint
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.base import ExecutableOption
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    InvalidCiphertextError,
)
from sqlmodel import Field, Relationship

from zenml.constants import TEXT_FIELD_MAX_LENGTH
from zenml.models import (
    SecretRequest,
    SecretResponse,
    SecretResponseBody,
    SecretResponseMetadata,
    SecretResponseResources,
    SecretUpdate,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import BaseSchema, NamedSchema
from zenml.zen_stores.schemas.schema_utils import (
    build_foreign_key_field,
    build_index,
)
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import jl_arg


class SecretDecodeError(Exception):
    """Raised when a secret cannot be decoded or decrypted."""


class SecretSchema(NamedSchema, table=True):
    """SQL Model for secrets.

    Attributes:
        name: The name of the secret.
        values: The values of the secret.
    """

    __tablename__ = "secret"
    __table_args__ = (
        UniqueConstraint(
            "name",
            "private",
            "user_id",
            name="unique_secret_name_private_scope_user",
        ),
    )

    private: bool

    internal: bool = Field(default=False)

    values: Optional[bytes] = Field(sa_column=Column(TEXT, nullable=True))

    user_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    user: "UserSchema" = Relationship(back_populates="secrets")

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = []

        if include_resources:
            options.extend([joinedload(jl_arg(SecretSchema.user))])

        return options

    @classmethod
    def _dump_secret_values(
        cls, values: Dict[str, str], encryption_engine: Optional[AesGcmEngine]
    ) -> bytes:
        """Dump the secret values to a string.

        Args:
            values: The secret values to dump.
            encryption_engine: The encryption engine to use to encrypt the
                secret values. If None, the values will be base64 encoded.

        Raises:
            ValueError: If the secret values do not fit in the database field.

        Returns:
            The serialized encrypted secret values.
        """
        serialized_values = json.dumps(values)

        if encryption_engine is None:
            encrypted_values = base64.b64encode(
                serialized_values.encode("utf-8")
            )
        else:
            encrypted_values = encryption_engine.encrypt(serialized_values)

        if len(encrypted_values) > TEXT_FIELD_MAX_LENGTH:
            raise ValueError(
                "Database representation of secret values exceeds max "
                "length. Please use fewer values or consider using shorter "
                "secret keys and/or values."
            )

        return encrypted_values

    @classmethod
    def _load_secret_values(
        cls,
        encrypted_values: bytes,
        encryption_engine: Optional[AesGcmEngine] = None,
    ) -> Dict[str, str]:
        """Load the secret values from a base64 encoded byte string.

        Args:
            encrypted_values: The serialized encrypted secret values.
            encryption_engine: The encryption engine to use to decrypt the
                secret values. If None, the values will be base64 decoded.

        Returns:
            The loaded secret values.

        Raises:
            SecretDecodeError: If the secret values cannot be decoded or
                decrypted.
        """
        if encryption_engine is None:
            try:
                serialized_values = base64.b64decode(encrypted_values).decode()
            except ValueError as e:
                raise SecretDecodeError(
                    "Could not decode base64 encoded secret values: {str(e)}"
                ) from e
        else:
            try:
                serialized_values = encryption_engine.decrypt(encrypted_values)
            except (ValueError, InvalidCiphertextError) as e:
                raise SecretDecodeError(
                    "Could not decrypt secret values. Please check that the "
                    f"encryption key is correct: {str(e)}"
                ) from e

        try:
            return cast(
                Dict[str, str],
                json.loads(serialized_values),
            )
        except json.JSONDecodeError as e:
            raise SecretDecodeError(
                "Could not decode secret values. Please check that the "
                f"secret values are valid JSON: {str(e)}"
            ) from e

    @classmethod
    def from_request(
        cls,
        secret: SecretRequest,
        internal: bool = False,
    ) -> "SecretSchema":
        """Create a `SecretSchema` from a `SecretRequest`.

        Args:
            secret: The `SecretRequest` from which to create the schema.
            internal: Whether the secret is internal.

        Returns:
            The created `SecretSchema`.
        """
        assert secret.user is not None, "User must be set for secret creation."
        return cls(
            name=secret.name,
            private=secret.private,
            user_id=secret.user,
            # Don't store secret values implicitly in the secret. The
            # SQL secret store will call `store_secret_values` to store the
            # values separately if SQL is used as the secrets store.
            values=None,
            internal=internal,
        )

    def update(
        self,
        secret_update: SecretUpdate,
    ) -> "SecretSchema":
        """Update a `SecretSchema` from a `SecretUpdate`.

        Args:
            secret_update: The `SecretUpdate` from which to update the schema.

        Returns:
            The updated `SecretSchema`.
        """
        # Don't update the secret values implicitly in the secret. The
        # SQL secret store will call `set_secret_values` to update the
        # values separately if SQL is used as the secrets store.
        for field, value in secret_update.model_dump(
            exclude_unset=True, exclude={"user", "values"}
        ).items():
            setattr(self, field, value)

        self.updated = utc_now()
        return self

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> SecretResponse:
        """Converts a secret schema to a secret model.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic


        Returns:
            The secret model.
        """
        metadata = None
        if include_metadata:
            metadata = SecretResponseMetadata()

        resources = None
        if include_resources:
            resources = SecretResponseResources(
                user=self.user.to_model() if self.user else None,
            )

        # Don't load the secret values implicitly in the secret. The
        # SQL secret store will call `get_secret_values` to load the
        # values separately if SQL is used as the secrets store.
        body = SecretResponseBody(
            user_id=self.user_id,
            created=self.created,
            updated=self.updated,
            private=self.private,
        )
        return SecretResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def get_secret_values(
        self,
        encryption_engine: Optional[AesGcmEngine] = None,
    ) -> Dict[str, str]:
        """Get the secret values for this secret.

        This method is used by the SQL secrets store to load the secret values
        from the database.

        Args:
            encryption_engine: The encryption engine to use to decrypt the
                secret values. If None, the values will be base64 decoded.

        Returns:
            The secret values

        Raises:
            KeyError: if no secret values for the given ID are stored in the
                secrets store.
        """
        if not self.values:
            raise KeyError(
                f"Secret values for secret {self.id} have not been stored in "
                f"the SQL secrets store."
            )
        return self._load_secret_values(self.values, encryption_engine)

    def set_secret_values(
        self,
        secret_values: Dict[str, str],
        encryption_engine: Optional[AesGcmEngine] = None,
    ) -> None:
        """Create a `SecretSchema` from a `SecretRequest`.

        This method is used by the SQL secrets store to store the secret values
        in the database.

        Args:
            secret_values: The new secret values.
            encryption_engine: The encryption engine to use to encrypt the
                secret values. If None, the values will be base64 encoded.
        """
        self.values = self._dump_secret_values(
            secret_values, encryption_engine
        )


class SecretResourceSchema(BaseSchema, table=True):
    """SQL Model for secret resource relationship."""

    __tablename__ = "secret_resource"
    __table_args__ = (
        build_index(
            table_name=__tablename__,
            column_names=[
                "resource_id",
                "resource_type",
                "secret_id",
            ],
            unique=True,
        ),
    )

    secret_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=SecretSchema.__tablename__,
        source_column="secret_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    resource_id: UUID
    resource_type: str = Field(sa_column=Column(VARCHAR(255), nullable=False))



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML Step HeartBeat functionality."""

import _thread
import contextvars
import logging
import threading
import time
from datetime import datetime, timezone
from typing import TYPE_CHECKING
from uuid import UUID

from zenml.enums import ExecutionStatus
from zenml.utils.time_utils import to_utc_timezone

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


class StepHeartBeatTerminationException(Exception):
    """Custom exception class for heartbeat termination."""

    pass


class StepHeartbeatWorker:
    """Worker class implementing heartbeat polling and remote termination."""

    STEP_HEARTBEAT_INTERVAL_SECONDS = (
        60 * 2
    )  # submit heartbeats every 2 minutes.

    MAX_HEARTBEAT_INTERVAL_SECONDS = 60 * 5
    MAX_CONSECUTIVE_FAILURES = 3

    def __init__(self, step_id: UUID):
        """Heartbeat worker constructor.

        Args:
            step_id: The step id heartbeat is running for.
        """
        self._step_id = step_id

        self._thread: threading.Thread | None = None
        self._running: bool = False
        self._terminated: bool = False

        self._consecutive_failures: int = 0
        self._heartbeat_interval_seconds: int = (
            self.STEP_HEARTBEAT_INTERVAL_SECONDS
        )

    # properties

    @property
    def is_running(self) -> bool:
        """Property function for running signal.

        Returns:
            True if the worker is running.
        """
        return self._running

    @property
    def is_terminated(self) -> bool:
        """Property function for termination signal.

        Returns:
            True if the worker has been terminated.
        """
        return self._terminated

    @property
    def interval(self) -> int:
        """Property function for heartbeat interval.

        Returns:
            The heartbeat polling interval value.
        """
        return self._heartbeat_interval_seconds

    @property
    def name(self) -> str:
        """Property function for heartbeat worker name.

        Returns:
            The name of the heartbeat worker.
        """
        return f"HeartBeatWorker-{self.step_id}"

    @property
    def step_id(self) -> UUID:
        """Property function for heartbeat worker step ID.

        Returns:
            The id of the step heartbeat is running for.
        """
        return self._step_id

    # public functions

    def start(self) -> None:
        """Start the heartbeat worker on a background thread."""
        if self._thread and self._thread.is_alive():
            return

        self._running = True
        self._terminated = False

        active_context = contextvars.copy_context()
        self._thread = threading.Thread(
            target=lambda: active_context.run(self._run),
            name=self.name,
            daemon=True,
        )
        self._thread.start()
        logger.debug(
            "Daemon thread %s started (interval=%s)", self.name, self.interval
        )

    def stop(self) -> None:
        """Stops the heartbeat worker."""
        if not self._running:
            return
        self._running = False
        logger.debug("%s stop requested", self.name)

    def is_alive(self) -> bool:
        """Liveness of the heartbeat worker thread.

        Returns:
            True if the heartbeat worker thread is alive, False otherwise.
        """
        t = self._thread
        return bool(t and t.is_alive())

    def _run(self) -> None:
        logger.debug("%s run() loop entered", self.name)
        try:
            while self._running:
                try:
                    self._heartbeat()
                    self._consecutive_failures = 0
                    # One-shot: signal the main thread and stop the loop.
                    if self._terminated:
                        logger.info(
                            "%s is remotely stopped, interrupting main thread",
                            self.name,
                        )
                        _thread.interrupt_main()  # raises KeyboardInterrupt in main thread
                        self._running = False
                    # Ensure we stop our own loop as well.

                except Exception as exc:
                    self._consecutive_failures += 1
                    # Log-and-continue policy for all other errors.
                    logger.debug(
                        "%s heartbeat() failed with %s", self.name, str(exc)
                    )

                    if (
                        self._consecutive_failures
                        == self.MAX_CONSECUTIVE_FAILURES
                    ):
                        logger.debug(
                            "Reached %s consecutive failures - applying increased heartbeat interval..",
                            self.MAX_CONSECUTIVE_FAILURES,
                        )
                        self._heartbeat_interval_seconds = (
                            self.MAX_HEARTBEAT_INTERVAL_SECONDS
                        )

                # Sleep after each attempt (even after errors, unless stopped).
                if self._running:
                    time.sleep(self.interval)
        finally:
            logger.debug("%s run() loop exiting", self.name)

    def _heartbeat(self) -> None:
        from zenml.config.global_config import GlobalConfiguration

        store = GlobalConfiguration().zen_store

        response = store.update_step_heartbeat(step_run_id=self.step_id)

        if not response.heartbeat_enabled:
            logger.debug("Heartbeat set to disabled - stopping worker")
            self.stop()
            return

        if response.pipeline_run_status in {
            ExecutionStatus.STOPPED,
            ExecutionStatus.STOPPING,
        }:
            self._terminated = True


def is_heartbeat_unhealthy(
    step_run_id: UUID,
    status: ExecutionStatus,
    latest_heartbeat: datetime | None,
    start_time: datetime | None = None,
    heartbeat_threshold: int | None = None,
) -> bool:
    """Utility function - Checks if step heartbeats indicate un-healthy execution.

    Args:
        step_run_id: The run step id.
        status: The run step status.
        latest_heartbeat: The run step latest heartbeat.
        start_time: The run step start time.
        heartbeat_threshold: If heartbeat enabled the max minutes without heartbeat
            for healthy, running tasks.

    Returns:
        True if the step heartbeat is unhealthy, False otherwise.
    """
    if not heartbeat_threshold:
        return False

    if status.is_finished:
        return False

    if latest_heartbeat:
        heartbeat_diff = datetime.now(tz=timezone.utc) - to_utc_timezone(
            latest_heartbeat
        )
    elif start_time:
        heartbeat_diff = datetime.now(tz=timezone.utc) - to_utc_timezone(
            start_time
        )
    else:
        return False

    logger.debug("Step %s heartbeat diff=%s", step_run_id, heartbeat_diff)

    if heartbeat_diff.total_seconds() > heartbeat_threshold * 60:
        return True

    return False



"""remove triggers [97109a4a8d26].

Revision ID: 97109a4a8d26
Revises: 8e36b8227160
Create Date: 2026-01-26 21:21:19.364737

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "97109a4a8d26"
down_revision = "0.93.3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_pipeline_run_trigger_execution_id_trigger_execution",
            type_="foreignkey",
        )
        batch_op.drop_column("trigger_execution_id")

    op.drop_table("trigger_execution")  # Should also drop the trigger fk
    op.drop_table("trigger")  # Should also drop the action, event_source fk
    op.drop_table("event_source")
    op.drop_table("action")


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "trigger_execution_id", sa.CHAR(length=32), nullable=True
            )
        )
        batch_op.create_foreign_key(
            "fk_pipeline_run_trigger_execution_id_trigger_execution",
            "trigger_execution",
            ["trigger_execution_id"],
            ["id"],
            ondelete="SET NULL",
        )

    op.create_table(
        "trigger_execution",
        sa.Column("id", sa.CHAR(length=32), nullable=False),
        sa.Column("created", sa.DATETIME(), nullable=False),
        sa.Column("updated", sa.DATETIME(), nullable=False),
        sa.Column("trigger_id", sa.CHAR(length=32), nullable=False),
        sa.Column("event_metadata", sa.BLOB(), nullable=True),
        sa.ForeignKeyConstraint(
            ["trigger_id"],
            ["trigger.id"],
            name="fk_trigger_execution_trigger_id_trigger",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "trigger",
        sa.Column("id", sa.CHAR(length=32), nullable=False),
        sa.Column("created", sa.DATETIME(), nullable=False),
        sa.Column("updated", sa.DATETIME(), nullable=False),
        sa.Column("name", sa.VARCHAR(), nullable=False),
        sa.Column("project_id", sa.CHAR(length=32), nullable=False),
        sa.Column("user_id", sa.CHAR(length=32), nullable=True),
        sa.Column("event_source_id", sa.CHAR(length=32), nullable=True),
        sa.Column("action_id", sa.CHAR(length=32), nullable=False),
        sa.Column("event_filter", sa.BLOB(), nullable=False),
        sa.Column("schedule", sa.BLOB(), nullable=True),
        sa.Column("description", sa.TEXT(), nullable=True),
        sa.Column("is_active", sa.BOOLEAN(), nullable=False),
        sa.ForeignKeyConstraint(
            ["action_id"],
            ["action.id"],
            name="fk_trigger_action_id_action",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["event_source_id"],
            ["event_source.id"],
            name="fk_trigger_event_source_id_event_source",
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_trigger_project_id_project",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_trigger_user_id_user",
            ondelete="SET NULL",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "name", "project_id", name="unique_trigger_name_in_project"
        ),
    )
    op.create_table(
        "action",
        sa.Column("id", sa.CHAR(length=32), nullable=False),
        sa.Column("created", sa.DATETIME(), nullable=False),
        sa.Column("updated", sa.DATETIME(), nullable=False),
        sa.Column("name", sa.VARCHAR(), nullable=False),
        sa.Column("project_id", sa.CHAR(length=32), nullable=False),
        sa.Column("user_id", sa.CHAR(length=32), nullable=True),
        sa.Column("service_account_id", sa.CHAR(length=32), nullable=False),
        sa.Column("auth_window", sa.INTEGER(), nullable=False),
        sa.Column("flavor", sa.VARCHAR(), nullable=False),
        sa.Column("plugin_subtype", sa.VARCHAR(), nullable=False),
        sa.Column("description", sa.TEXT(), nullable=True),
        sa.Column("configuration", sa.BLOB(), nullable=False),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_action_project_id_project",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["service_account_id"],
            ["user.id"],
            name="fk_action_service_account_id_user",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_action_user_id_user",
            ondelete="SET NULL",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "name", "project_id", name="unique_action_name_in_project"
        ),
    )
    op.create_table(
        "event_source",
        sa.Column("id", sa.CHAR(length=32), nullable=False),
        sa.Column("created", sa.DATETIME(), nullable=False),
        sa.Column("updated", sa.DATETIME(), nullable=False),
        sa.Column("name", sa.VARCHAR(), nullable=False),
        sa.Column("project_id", sa.CHAR(length=32), nullable=False),
        sa.Column("user_id", sa.CHAR(length=32), nullable=True),
        sa.Column("flavor", sa.VARCHAR(), nullable=False),
        sa.Column("plugin_subtype", sa.VARCHAR(), nullable=False),
        sa.Column("description", sa.TEXT(), nullable=True),
        sa.Column("configuration", sa.BLOB(), nullable=False),
        sa.Column("is_active", sa.BOOLEAN(), nullable=False),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_event_source_project_id_project",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_event_source_user_id_user",
            ondelete="SET NULL",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "name", "project_id", name="unique_event_source_name_in_project"
        ),
    )
    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for building manifests for k8s pods."""

import base64
import os
import sys
from typing import Any, Dict, List, Mapping, Optional

from kubernetes import client as k8s_client

from zenml.client import Client
from zenml.config.global_config import GlobalConfiguration
from zenml.constants import ENV_ZENML_ENABLE_REPO_INIT_WARNINGS
from zenml.integrations.airflow.orchestrators.dag_generator import (
    ENV_ZENML_LOCAL_STORES_PATH,
)
from zenml.integrations.kubernetes.pod_settings import KubernetesPodSettings
from zenml.logger import get_logger
from zenml.utils import string_utils

logger = get_logger(__name__)


def add_local_stores_mount(
    pod_spec: k8s_client.V1PodSpec,
) -> None:
    """Makes changes in place to the configuration of the pod spec.

    Configures mounted volumes for stack components that write to a local
    path.

    Args:
        pod_spec: The pod spec to update.
    """
    assert len(pod_spec.containers) == 1
    container_spec: k8s_client.V1Container = pod_spec.containers[0]

    stack = Client().active_stack

    stack.check_local_paths()

    local_stores_path = GlobalConfiguration().local_stores_path

    host_path = k8s_client.V1HostPathVolumeSource(
        path=local_stores_path, type="Directory"
    )

    pod_spec.volumes = pod_spec.volumes or []
    pod_spec.volumes.append(
        k8s_client.V1Volume(
            name="local-stores",
            host_path=host_path,
        )
    )
    container_spec.volume_mounts = container_spec.volume_mounts or []
    container_spec.volume_mounts.append(
        k8s_client.V1VolumeMount(
            name="local-stores",
            mount_path=local_stores_path,
        )
    )

    if sys.platform == "win32":
        # File permissions are not checked on Windows. This if clause
        # prevents mypy from complaining about unused 'type: ignore'
        # statements
        pass
    else:
        # Run KFP containers in the context of the local UID/GID
        # to ensure that the local stores can be shared
        # with the local pipeline runs.
        pod_spec.security_context = k8s_client.V1SecurityContext(
            run_as_user=os.getuid(),
            run_as_group=os.getgid(),
        )

    container_spec.env = container_spec.env or []
    container_spec.env.append(
        k8s_client.V1EnvVar(
            name=ENV_ZENML_LOCAL_STORES_PATH,
            value=local_stores_path,
        )
    )


def build_pod_manifest(
    pod_name: Optional[str],
    image_name: str,
    command: List[str],
    args: List[str],
    privileged: bool,
    pod_settings: Optional[KubernetesPodSettings] = None,
    service_account_name: Optional[str] = None,
    env: Optional[Dict[str, str]] = None,
    labels: Optional[Dict[str, str]] = None,
    mount_local_stores: bool = False,
    owner_references: Optional[List[k8s_client.V1OwnerReference]] = None,
    termination_grace_period_seconds: Optional[int] = 30,
) -> k8s_client.V1Pod:
    """Build a Kubernetes pod manifest for a ZenML run or step.

    Args:
        pod_name: Name of the pod.
        image_name: Name of the Docker image.
        command: Command to execute the entrypoint in the pod.
        args: Arguments provided to the entrypoint command.
        privileged: Whether to run the container in privileged mode.
        pod_settings: Optional settings for the pod.
        service_account_name: Optional name of a service account.
            Can be used to assign certain roles to a pod, e.g., to allow it to
            run Kubernetes commands from within the cluster.
        env: Environment variables to set.
        labels: Labels to add to the pod.
        mount_local_stores: Whether to mount the local stores path inside the
            pod.
        owner_references: List of owner references for the pod.
        termination_grace_period_seconds: The amount of seconds to wait for a
            pod to shutdown gracefully.

    Returns:
        Pod manifest.
    """
    env = env.copy() if env else {}
    env.setdefault(ENV_ZENML_ENABLE_REPO_INIT_WARNINGS, "False")

    security_context = k8s_client.V1SecurityContext(privileged=privileged)
    container_spec = k8s_client.V1Container(
        name="main",
        image=image_name,
        command=command,
        args=args,
        env=[
            k8s_client.V1EnvVar(name=name, value=value)
            for name, value in env.items()
        ],
        security_context=security_context,
    )
    image_pull_secrets = []
    if pod_settings:
        image_pull_secrets = [
            k8s_client.V1LocalObjectReference(name=name)
            for name in pod_settings.image_pull_secrets
        ]

    pod_spec = k8s_client.V1PodSpec(
        containers=[container_spec],
        restart_policy="Never",
        image_pull_secrets=image_pull_secrets,
        termination_grace_period_seconds=termination_grace_period_seconds,
    )

    if service_account_name is not None:
        pod_spec.service_account_name = service_account_name

    # Apply pod settings if provided
    labels = labels or {}

    if pod_settings:
        substitutions = {
            "{{ image }}": image_name,
        }
        add_pod_settings(
            pod_spec=pod_spec,
            settings=pod_settings,
            substitutions=substitutions,
        )

    if pod_settings and pod_settings.labels:
        labels.update(pod_settings.labels)

    pod_metadata = k8s_client.V1ObjectMeta(
        name=pod_name,
        labels=labels,
        owner_references=owner_references,
    )

    if pod_settings and pod_settings.annotations:
        pod_metadata.annotations = pod_settings.annotations

    pod_manifest = k8s_client.V1Pod(
        kind="Pod",
        api_version="v1",
        metadata=pod_metadata,
        spec=pod_spec,
    )

    if mount_local_stores:
        add_local_stores_mount(pod_spec)

    return pod_manifest


def add_pod_settings(
    pod_spec: k8s_client.V1PodSpec,
    settings: KubernetesPodSettings,
    substitutions: Dict[str, str],
) -> None:
    """Updates pod `spec` fields in place if passed in orchestrator settings.

    Args:
        pod_spec: Pod spec to update.
        settings: Pod settings to apply.
        substitutions: Substitutions to apply to additional pod spec arguments.
    """
    if settings.node_selectors:
        pod_spec.node_selector = settings.node_selectors

    if settings.affinity:
        pod_spec.affinity = settings.affinity

    if settings.tolerations:
        pod_spec.tolerations = settings.tolerations

    for container in pod_spec.containers:
        assert isinstance(container, k8s_client.V1Container)
        container._resources = settings.resources
        if settings.volume_mounts:
            if container.volume_mounts:
                container.volume_mounts.extend(settings.volume_mounts)
            else:
                container.volume_mounts = settings.volume_mounts

        if settings.env:
            if container.env:
                container.env.extend(settings.env)
            else:
                container.env = settings.env

        if settings.env_from:
            if container.env_from:
                container.env_from.extend(settings.env_from)
            else:
                container.env_from = settings.env_from

        if settings.container_security_context:
            if container.security_context:
                existing_dict = container.security_context.to_dict()
                existing_dict.update(settings.container_security_context)
                container.security_context = existing_dict
            else:
                container.security_context = (
                    settings.container_security_context
                )

    if settings.volumes:
        if pod_spec.volumes:
            pod_spec.volumes.extend(settings.volumes)
        else:
            pod_spec.volumes = settings.volumes

    if settings.host_ipc:
        pod_spec.host_ipc = settings.host_ipc

    if settings.scheduler_name:
        pod_spec.scheduler_name = settings.scheduler_name

    for key, value in settings.additional_pod_spec_args.items():
        if not hasattr(pod_spec, key):
            logger.warning(f"Ignoring invalid Pod Spec argument `{key}`.")
        else:
            if value is None:
                continue

            def _substitute(v: str) -> str:
                return substitutions.get(v, v)

            value = string_utils.substitute_string(value, _substitute)

            existing_value = getattr(pod_spec, key)
            if isinstance(existing_value, list):
                existing_value.extend(value)
            elif isinstance(existing_value, dict):
                existing_value.update(value)
            else:
                setattr(pod_spec, key, value)


def build_role_binding_manifest_for_service_account(
    name: str,
    role_name: str,
    service_account_name: str,
    namespace: str = "default",
) -> Dict[str, Any]:
    """Build a manifest for a role binding of a service account.

    Args:
        name: Name of the cluster role binding.
        role_name: Name of the role.
        service_account_name: Name of the service account.
        namespace: Kubernetes namespace. Defaults to "default".

    Returns:
        Manifest for a cluster role binding of a service account.
    """
    return {
        "apiVersion": "rbac.authorization.k8s.io/v1",
        "kind": "RoleBinding",
        "metadata": {"name": name},
        "subjects": [
            {
                "kind": "ServiceAccount",
                "name": service_account_name,
                "namespace": namespace,
            }
        ],
        "roleRef": {
            "kind": "ClusterRole",
            "name": role_name,
            "apiGroup": "rbac.authorization.k8s.io",
        },
    }


def build_service_account_manifest(
    name: str, namespace: str = "default"
) -> Dict[str, Any]:
    """Build the manifest for a service account.

    Args:
        name: Name of the service account.
        namespace: Kubernetes namespace. Defaults to "default".

    Returns:
        Manifest for a service account.
    """
    return {
        "apiVersion": "v1",
        "metadata": {
            "name": name,
            "namespace": namespace,
        },
    }


def build_namespace_manifest(
    namespace: str, labels: Optional[Dict[str, str]] = None
) -> Dict[str, Any]:
    """Build the manifest for a new namespace.

    Args:
        namespace: Kubernetes namespace.
        labels: Optional labels to apply to the namespace. These labels are used
            to track namespace ownership for cleanup. For example, adding
            'zenml-deployment-id' allows us to identify and safely delete
            namespaces created by ZenML deployments.

    Returns:
        Manifest of the new namespace.
    """
    metadata: Dict[str, Any] = {"name": namespace}
    if labels:
        metadata["labels"] = labels

    return {
        "apiVersion": "v1",
        "kind": "Namespace",
        "metadata": metadata,
    }


def build_secret_manifest(
    name: str,
    data: Mapping[str, Optional[str]],
    namespace: Optional[str] = None,
    secret_type: str = "Opaque",
) -> Dict[str, Any]:
    """Builds a Kubernetes secret manifest.

    Args:
        name: Name of the secret.
        data: The secret data.
        namespace: The namespace to create the secret in.
        secret_type: The secret type.

    Returns:
        The secret manifest.
    """
    encoded_data = {
        key: base64.b64encode(value.encode()).decode() if value else None
        for key, value in data.items()
    }

    return {
        "apiVersion": "v1",
        "kind": "Secret",
        "metadata": {"name": name, "namespace": namespace},
        "type": secret_type,
        "data": encoded_data,
    }


def pod_template_manifest_from_pod(
    pod: k8s_client.V1Pod,
) -> k8s_client.V1PodTemplateSpec:
    """Build a Kubernetes pod template manifest from a pod.

    Args:
        pod: The pod manifest to build the template from.

    Returns:
        The pod template manifest.
    """
    return k8s_client.V1PodTemplateSpec(
        metadata=pod.metadata,
        spec=pod.spec,
    )


def build_job_manifest(
    job_name: str,
    pod_template: k8s_client.V1PodTemplateSpec,
    backoff_limit: Optional[int] = None,
    ttl_seconds_after_finished: Optional[int] = None,
    labels: Optional[Dict[str, str]] = None,
    annotations: Optional[Dict[str, str]] = None,
    active_deadline_seconds: Optional[int] = None,
    pod_failure_policy: Optional[Dict[str, Any]] = None,
    owner_references: Optional[List[k8s_client.V1OwnerReference]] = None,
) -> k8s_client.V1Job:
    """Build a Kubernetes job manifest.

    Args:
        job_name: Name of the job.
        pod_template: The pod template to use for the job.
        backoff_limit: The backoff limit for the job.
        ttl_seconds_after_finished: The TTL seconds after finished for the job.
        labels: The labels to use for the job.
        annotations: The annotations to use for the job.
        active_deadline_seconds: The active deadline seconds for the job.
        pod_failure_policy: The pod failure policy for the job.
        owner_references: The owner references for the job.

    Returns:
        The Kubernetes job manifest.
    """
    job_spec = k8s_client.V1JobSpec(
        template=pod_template,
        backoff_limit=backoff_limit,
        parallelism=1,
        ttl_seconds_after_finished=ttl_seconds_after_finished,
        active_deadline_seconds=active_deadline_seconds,
        pod_failure_policy=pod_failure_policy,
    )

    job_metadata = k8s_client.V1ObjectMeta(
        name=job_name,
        labels=labels,
        annotations=annotations,
        owner_references=owner_references,
    )

    return k8s_client.V1Job(spec=job_spec, metadata=job_metadata)


def job_template_manifest_from_job(
    job: k8s_client.V1Job,
) -> k8s_client.V1JobTemplateSpec:
    """Build a Kubernetes job template manifest from a job.

    Args:
        job: The job manifest to build the template from.

    Returns:
        The job template manifest.
    """
    return k8s_client.V1JobTemplateSpec(
        metadata=job.metadata,
        spec=job.spec,
    )


def build_cron_job_manifest(
    job_template: k8s_client.V1JobTemplateSpec,
    cron_expression: str,
    successful_jobs_history_limit: Optional[int] = None,
    failed_jobs_history_limit: Optional[int] = None,
    concurrency_policy: Optional[str] = None,
    starting_deadline_seconds: Optional[int] = None,
) -> k8s_client.V1CronJob:
    """Build a Kubernetes cron job manifest.

    Args:
        job_template: The job template to use for the cron job.
        cron_expression: The cron expression to use for the cron job.
        successful_jobs_history_limit: The number of successful jobs to keep.
        failed_jobs_history_limit: The number of failed jobs to keep.
        concurrency_policy: The concurrency policy for the cron job
            (Allow, Forbid, or Replace).
        starting_deadline_seconds: The deadline in seconds for starting
            a job if it misses its scheduled time.

    Returns:
        The Kubernetes cron job manifest.
    """
    spec = k8s_client.V1CronJobSpec(
        schedule=cron_expression,
        successful_jobs_history_limit=successful_jobs_history_limit,
        failed_jobs_history_limit=failed_jobs_history_limit,
        concurrency_policy=concurrency_policy,
        starting_deadline_seconds=starting_deadline_seconds,
        job_template=job_template,
    )

    return k8s_client.V1CronJob(
        kind="CronJob",
        api_version="batch/v1",
        metadata=job_template.metadata,
        spec=spec,
    )
